from typing import List
from typing import Union

import numpy as np
import pandas as pd

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import GeneralizedIDs
from sonusai.mixture.types import Predict
from sonusai.mixture.types import Truth


def class_summary(mixdb: MixtureDatabase,
                  mixids: GeneralizedIDs,
                  truth_f: Truth,
                  predict: Predict,
                  pred_thr: Union[float, List[float]] = 0,
                  truth_thr: float = 0.5,
                  timesteps: int = 0) -> pd.DataFrame:
    """ Calculate table of metrics per class, and averages for a list
        of mixtures using truth and prediction data [features, num_classes]
        Example:
        Generate multi-class metric summary into table, for example:
                        PPV     TPR      F1     FPR     ACC   AP  AUC  Support
            Class 1     0.71    0.80    0.75    0.00    0.99            44
            Class 2     0.90    0.76    0.82    0.00    0.99            128
            Class 3     0.86    0.82    0.84    0.04    0.93            789
            Other       0.94    0.96    0.95    0.18    0.92            2807

          micro-avg                     0.92    0.027                   3768
          macro avg     0.85    0.83    0.84    0.05    0.96            3768
          micro-avgwo
    """
    from sonusai.metrics import one_hot

    num_classes = truth_f.shape[1]

    y_truth_f, y_predict = get_mixids_data(mixdb, mixids, truth_f, predict)

    if not mixdb.truth_mutex and num_classes > 1:
        if np.ndim(pred_thr) == 0 and pred_thr == 0:
            pred_thr = 0.5

        if np.ndim(pred_thr) == 1 and len(pred_thr) == 1:
            if pred_thr[0] == 0:
                pred_thr = 0.5

    _, metrics, _, _, _, metavg = one_hot(y_truth_f, y_predict, pred_thr, truth_thr, timesteps)

    # [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]
    table_idx = np.array([2, 1, 6, 4, 0, 12, 13, 9])
    col_n = ['PPV', 'TPR', 'F1', 'FPR', 'ACC', 'AP', 'AUC', 'Support']
    if mixdb.truth_mutex:
        if len(mixdb.class_labels) >= num_classes - 1:  # labels exist with or without Other
            row_n = mixdb.class_labels
            if len(mixdb.class_labels) == num_classes - 1:  # Other label does not exist, so add it
                row_n.append('Other')
        else:
            row_n = ([f'Class {i}' for i in range(1, num_classes)])
            row_n.append('Other')
    else:
        if len(mixdb.class_labels) == num_classes:
            row_n = mixdb.class_labels
        else:
            row_n = ([f'Class {i}' for i in range(1, num_classes + 1)])

    df = pd.DataFrame(metrics[:, table_idx], columns=col_n, index=row_n)

    # [miPPV, miTPR, miF1, miFPR, miACC, miAP, miAUC, TPSUM]
    avg_row_n = ['Macro-avg', 'Micro-avg', 'Weighted-avg']
    dfavg = pd.DataFrame(metavg, columns=col_n, index=avg_row_n)

    # dfblank = pd.DataFrame([''])
    # pd.concat([df, dfblank, dfblank, dfavg])

    classdf = pd.concat([df, dfavg])
    # classdf = classdf.round(2)
    classdf['Support'] = classdf['Support'].astype(int)

    return classdf



from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import ClassCount
from sonusai.mixture.types import GeneralizedIDs


def get_class_count_from_mixids(mixdb: MixtureDatabase, mixids: GeneralizedIDs = None) -> ClassCount:
    """ Sums the class counts for given mixids
    """
    from sonusai import SonusAIError

    total_class_count = [0] * mixdb.num_classes
    mixids = mixdb.mixids_to_list(mixids)
    for mixid in mixids:
        class_count = mixdb.mixture_class_count(mixid)
        for cl in range(mixdb.num_classes):
            total_class_count[cl] += class_count[cl]

    if mixdb.truth_mutex:
        # Compute the class count for the 'other' class
        if total_class_count[-1] != 0:
            raise SonusAIError('Error: truth_mutex was set, but the class count for the last count was non-zero.')
        total_class_count[-1] = sum([mixdb.mixture_samples(mixid) for mixid in mixids]) - sum(total_class_count)

    return total_class_count



import numpy as np


def stack_complex(unstacked: np.ndarray) -> np.ndarray:
    """Stack a complex array

    A stacked array doubles the last dimension and organizes the data as:
        - first half is all the real data
        - second half is all the imaginary data

    :param unstacked: An nD array (n > 1) containing complex data
    :return: A stacked array
    :raises TypeError:
    """
    if not unstacked.ndim > 1:
        raise TypeError('unstacked must have more than 1 dimension')

    shape = list(unstacked.shape)
    shape[-1] = shape[-1] * 2
    stacked = np.empty(shape, dtype=np.complex64)
    half = unstacked.shape[-1]
    stacked[..., :half] = np.real(unstacked)
    stacked[..., half:] = np.imag(unstacked)

    return stacked


def unstack_complex(stacked: np.ndarray) -> np.ndarray:
    """Unstack a stacked complex array

    :param stacked: An nD array (n > 1) where the last dimension contains stacked complex data in which the first half
        is all the real data and the second half is all the imaginary data
    :return: An unstacked complex array
    :raises TypeError:
    """
    if not stacked.ndim > 1:
        raise TypeError('stacked must have more than 1 dimension')

    if stacked.shape[-1] % 2 != 0:
        raise TypeError('last dimension of stacked must be a multiple of 2')

    half = stacked.shape[-1] // 2
    unstacked = 1j * stacked[..., half:]
    unstacked += stacked[..., :half]

    return unstacked


def stacked_complex_real(stacked: np.ndarray) -> np.ndarray:
    """Get the real elements from a stacked complex array

    :param stacked: An nD array (n > 1) where the last dimension contains stacked complex data in which the first half
        is all the real data and the second half is all the imaginary data
    :return: The real elements
    :raises TypeError:
    """
    if not stacked.ndim > 1:
        raise TypeError('stacked must have more than 1 dimension')

    if stacked.shape[-1] % 2 != 0:
        raise TypeError('last dimension of stacked must be a multiple of 2')

    half = stacked.shape[-1] // 2
    return stacked[..., :half]


def stacked_complex_imag(stacked: np.ndarray) -> np.ndarray:
    """Get the imaginary elements from a stacked complex array

    :param stacked: An nD array (n > 1) where the last dimension contains stacked complex data in which the first half
        is all the real data and the second half is all the imaginary data
    :return: The imaginary elements
    :raises TypeError:
    """
    if not stacked.ndim > 1:
        raise TypeError('stacked must have more than 1 dimension')

    if stacked.shape[-1] % 2 != 0:
        raise TypeError('last dimension of stacked must be a multiple of 2')

    half = stacked.shape[-1] // 2
    return stacked[..., half:]



from typing import Union

import numpy as np

from sonusai.mixture.types import Predict
from sonusai.mixture.types import Truth


def one_hot(truth: Truth,
            predict: Predict,
            predict_thr: Union[float, np.ndarray] = 0,
            truth_thr: float = 0.5,
            timesteps: int = -1) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    """ Calculates metrics from one-hot prediction and truth data (numpy float arrays) where
        both are one-hot probabilities (or quantized decisions) for each class
        with size [frames, num_classes] or [frames, timesteps, num_classes].
        For metrics that require it, truth and pred decisions will be made using threshold >= predict_thr.
        Some metrics like AP and AUC do not depend on predict_thr for predict, but still use truth >= predict_thr

        predict_thr sets the decision threshold(s) applied to predict data for some metrics, thus allowing
        the input to be continuous probabilities, for AUC-type metrics and root-mean-square error (rmse).
        1. Default = 0 (multiclass or binary) which infers:
           binary (num_classes = 1)         use >= 0.5 for truth and pred (same as argmax() for binary)
           multi-class/single-label    if truth_mutex= = true, use argmax() used on both truth and pred
           note multi-label metrics are disabled for predict_thr = 0, must set predict_thr > 0

        2. predict_thr > 0 (multilabel or binary) scalar or a vector [num_classes, 1] then use
           predict_thr as a binary decision threshold in each class:
           binary (num_classes = 1)         use >= predict_thr[0] for pred and predict_thr[num_classes+1] for truth
                                       if it exists, else use >= 0.5 for truth
           multi-label                 use >= predict_thr for pred if scalar, or predict_thr[class_idx] if vector
                                       use >= predict_thr[num_classes+1] for truth if exists, else 0.5
           note multi-class/single-label inputs are meaningless in this mode, use predict_thr = 0 argmax mode

        num_classes is inferred from 1D, 2D, or 3D truth inputs by default (default timesteps = -1 which implies None).
        Only set timesteps > 0 in case of ambiguous binary 2D case where input [frames, timesteps],
        then it must set to the number of timesteps (which will be > 0).
        It is safe to always set timesteps <= 0 for binary inputs, and if truth.shape[2] exists

        returns metrics over all frames + timesteps:
        mcm     [num_classes, 2, 2]            multiclass confusion matrix count ove
        metrics [num_classes, 14]              [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]
        cm      [num_classes, num_classes]     confusion matrix
        cmn     [num_classes, num_classes]     normalized confusion matrix
        rmse    [num_classes, 1]               RMS error over all frames + timesteps, before threshold decision
        mavg    [3, 8] averages                macro, micro, weighted [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
    """
    import warnings

    from sklearn.metrics import average_precision_score
    from sklearn.metrics import confusion_matrix
    from sklearn.metrics import multilabel_confusion_matrix
    from sklearn.metrics import precision_recall_fscore_support
    from sklearn.metrics import roc_auc_score

    from sonusai import SonusAIError
    from sonusai.utils import get_num_classes_from_predict
    from sonusai.utils import reshape_outputs

    if truth.shape != predict.shape:
        raise SonusAIError('truth and predict are not the same shape')

    predict, truth = reshape_outputs(predict=predict, truth=truth, timesteps=timesteps)
    num_classes = get_num_classes_from_predict(predict=predict, timesteps=timesteps)

    # Regression metric root-mean-square-error always works
    rmse = np.sqrt(np.mean(np.square(truth - predict), axis=0))

    # Calculate default predict decision thresholds based on mode
    if not np.ndim(predict_thr) and predict_thr == 0:
        # if scalar and 0, set defaults
        if num_classes == 1:
            # binary case default >= 0.5 which is equiv to argmax()
            predict_thr = np.atleast_1d(0.5)
        else:
            # multiclass, single-label (argmax mode)
            predict_thr = np.atleast_1d(0)
    else:
        if np.ndim(predict_thr) > 1:
            # multi-label with custom thr vector
            assert predict_thr.shape[0] == num_classes
        else:
            if predict_thr == 0:
                # binary or multi-label scalar default
                predict_thr = np.atleast_1d(0.5)
            else:
                # user specified binary or multi-label scalar
                predict_thr = np.atleast_1d(predict_thr)

    # Convert continuous probabilities to binary via argmax() or threshold comparison
    # and create labels of int encoded (0:num_classes-1), and then equivalent one-hot
    if num_classes == 1:  # If binary
        labels = ([i for i in range(0, 2)])  # int encoded 0,1
        plabel = np.int8(predict >= predict_thr)  # [frames, 1], default 0.5 is equiv. to argmax()
        tlabel = np.int8(truth >= truth_thr)  # [frames, 1]
        predb = plabel
        truthb = tlabel
    else:
        labels = ([i for i in range(0, num_classes)])  # int encoded 0,...,num_classes-1
        if predict_thr[0] == 0:  # multiclass single-label (mutex), use argmax
            plabel = np.argmax(predict, axis=-1)  # [frames, 1] labels
            tlabel = np.argmax(truth, axis=-1)  # [frames, 1] labels
            predb = np.zeros(predict.shape, dtype=np.int8)  # [frames, num_classes] one-hot binary
            truthb = np.zeros(truth.shape, dtype=np.int8)  # [frames, num_classes] one-hot binary
            predb[np.arange(predb.shape[0]), plabel] = 1  # single-label [frames, num_classes]
            if np.sum(truth):  # special case all zero truth leave tlabel all zeros
                truthb[np.arange(truthb.shape[0]), tlabel] = 1  # single-label [frames, num_classes]
        else:  # multi-label prob threshold comparison (multiple classes)
            predb = np.int8(predict >= predict_thr.transpose())  # [frames, num_classes] multilabel one-hot decision
            truthb = np.int8(truth >= truth_thr)  # [frames, num_classes] multilabel one-hot decision
            # Return argmax() for optional single-label confusion matrix metrics
            plabel = np.argmax(predict, axis=-1)  # [frames, 1] labels
            tlabel = np.argmax(truth, axis=-1)  # [frames, 1] labels

    # debug checks to understand ap, auc:
    # from sklearn.metrics import roc_curve
    # fpr, tpr, thr = roc_curve(truthb[:,0],predict[:,0],drop_intermediate=False)
    # from sklearn.metrics import precision_recall_curve
    # precision, recall, thr  = precision_recall_curve(truthb[:,0], predict[:,0])
    # from sklearn.metrics import RocCurveDisplay
    # RocCurveDisplay.from_predictions(truthb[:,0],predict[:,0])  # Plot ROC class0

    # Create [num_classes, 2, 2] multi-label confusion matrix (mcm)
    # Note - must include labels or sklearn func. will omit non-exiting classes
    mcm = multilabel_confusion_matrix(truthb, predb, labels=labels)

    if num_classes == 1:
        mcm = mcm[1:]  # remove dim 0 if binary

    # Create [num_classes, num_classes] normalized confusion matrix
    cmn = confusion_matrix(tlabel, plabel, labels=labels, normalize='true')

    # Create [num_classes, num_classes] confusion matrix
    cm = confusion_matrix(tlabel, plabel, labels=labels)

    # Combine all per-class metrics into a single array
    # [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]
    metrics = np.zeros((num_classes, 14))
    # threshold_optpr = np.zeros((num_classes, 1))
    eps = np.finfo(float).eps
    for nci in range(num_classes):
        # True negative
        TN = mcm[nci, 0, 0]
        # False positive
        FP = mcm[nci, 0, 1]
        # False negative
        FN = mcm[nci, 1, 0]
        # True positive
        TP = mcm[nci, 1, 1]
        # Accuracy
        ACC = (TP + TN) / (TP + TN + FP + FN + eps)
        # True positive rate, sensitivity, recall, hit rate (note eps in numerator)
        # When ``true positive + false negative == 0``, recall is undefined, set to 0
        TPR = TP / (TP + FN + eps)
        # Precision, positive predictive value
        # When ``true positive + false positive == 0``, precision is undefined, set to 0
        PPV = TP / (TP + FP + eps)
        # Specificity i.e., selectivity, or true negative rate
        TNR = TN / (TN + FP + eps)
        # False positive rate = 1-specificity, roc x-axis
        FPR = FP / (TN + FP + eps)
        # HitFA used by some separation research, close match to MCC
        HITFA = TPR - FPR
        # F1 harmonic mean of precision, recall = 2*PPV*TPR / (PPV + TPR)
        F1 = 2 * TP / (2 * TP + FP + FN + eps)
        # Matthew correlation coefficient
        MCC = (TP * TN - FP * FN) / (np.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)) + eps)
        # Num. negatives total (truth), also = TN+FP denom of FPR
        NT = sum(mcm[nci, 0, :])
        # Num. positives total (truth), also = FN+TP denom of TPR, precision
        PT = sum(mcm[nci, 1, :])
        # Average Precision also called area under the PR curve AUCPR and
        # AUC ROC curve using binary-ized truth and continuous prediction probabilities
        # sklearn returns nan if no active truth in a class but w/un-suppressible div-by-zero warning
        if np.sum(truthb[:, nci]) == 0:  # if no active classes both sklearn will fail, set to NaN
            AUC = np.NaN
            AP = np.NaN
            # threshold_optpr[nci] = np.NaN
        else:
            AP = average_precision_score(truthb[:, nci], predict[:, nci], average=None)
            if len(np.unique(truthb[:, nci])) < 2:  # if active classes not > 1 AUC must be NaN
                AUC = np.NaN  # i.e. all ones sklearn auc will fail
            else:
                AUC = roc_auc_score(truthb[:, nci], predict[:, nci], average=None)
            # # Optimal threshold from PR curve, optimizes f-score
            # precision, recall, thresholds = precision_recall_curve(truthb[:, nci], predict[:, nci])
            # fscore = (2 * precision * recall) / (precision + recall)
            # ix = np.argmax(fscore)  # index of largest f1 score
            # threshold_optpr[nci] = thresholds[ix]

        metrics[nci, :] = [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]

    # Calculate averages into single array, 3 types for now Macro, Micro, Weighted
    mavg = np.zeros((3, 8), dtype=np.float32)
    s = np.sum(metrics[:, 9].astype(int))  # support = sum (true pos total = FN+TP ) over classes

    # macro average [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
    with warnings.catch_warnings():
        warnings.filterwarnings(action='ignore', message='Mean of empty slice')
        mavg[0, :] = [np.mean(metrics[:, 2]), np.mean(metrics[:, 1]), np.mean(metrics[:, 6]), np.mean(metrics[:, 4]),
                      np.mean(metrics[:, 0]), np.nanmean(metrics[:, 12]), np.nanmean(metrics[:, 13]), s]

    # micro average, micro-F1 = micro-precision = micro-recall = accuracy
    if num_classes > 1:
        tp_sum = np.sum(metrics[:, 10])  # TP all classes
        rm = tp_sum / (np.sum(metrics[:, 9]) + eps)  # micro mean PPV = TP / (PT=FN+TP)
        fp_sum = np.sum(metrics[:, 11])  # FP false-positives all classes
        fpm = fp_sum / (np.sum(metrics[:, 8]) + eps)  # micro mean FPR = FP / (NT=TN+FP)
        pm = tp_sum / (tp_sum + fp_sum + eps)  # micro mean TPR = TP / (TP+FP) (note: same as rm for micro-avg)
        fn_sum = sum(mcm[:, 1, 0])
        f1m = 2 * tp_sum / (2 * tp_sum + fp_sum + fn_sum + eps)
        tn_sum = sum(mcm[:, 0, 0])
        accm = (tp_sum + tn_sum) / (tp_sum + tn_sum + fp_sum + fn_sum + eps)
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore', message='invalid value encountered in true_divide')
            miap = average_precision_score(truthb, predict, average='micro')
        if np.sum(truthb):  # no activity over all classes
            miauc = roc_auc_score(truthb, predict, average='micro')
        else:
            miauc = np.NaN

        # [miPPV, miTPR, miF1, miFPR, miACC, miAP, miAUC, TPSUM]
        mavg[1, :] = [pm, rm, f1m, fpm, accm, miap, miauc, s]  # specific format, last 3 are unique

        # weighted average TBD
        wp, wr, wf1, _ = precision_recall_fscore_support(truthb, predb, average='weighted', zero_division=0)
        if np.sum(truthb):
            taidx = np.sum(truthb, axis=0) > 0
            wap = average_precision_score(truthb[:, taidx], predict[:, taidx], average='weighted')
            if len(np.unique(truthb[:, taidx])) < 2:
                wauc = np.NaN
            else:
                wauc = roc_auc_score(truthb[:, taidx], predict[:, taidx], average='weighted')
        else:
            wap = np.NaN
            wauc = np.NaN

        mavg[2, :] = [wp, wr, wf1, 0, 0, wap, wauc, s]
    else:  # binary case, all are same
        mavg[1, :] = mavg[0, :]
        mavg[2, :] = mavg[0, :]

    return mcm, metrics, cm, cmn, rmse, mavg



from typing import Union

import numpy as np
import pandas as pd

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import GeneralizedIDs
from sonusai.mixture.types import Predict
from sonusai.mixture.types import Segsnr
from sonusai.mixture.types import Truth


def snr_summary(mixdb: MixtureDatabase,
                mixid: GeneralizedIDs,
                truth_f: Truth,
                predict: Predict,
                segsnr: Segsnr = None,
                predict_thr: Union[float, np.ndarray] = 0,
                truth_thr: float = 0.5,
                timesteps: int = 0) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame, dict):
    """Calculate average-over-class metrics per SNR over specified mixture list.
       Inputs:
         mixdb        Mixture database
         mixid
         truth_f      Truth/labels [features, num_classes]
         predict      Prediction data / neural net model one-hot out [features, num_classes]
         segsnr       Segmental SNR from SonusAI genft  [transform_frames, 1]
         predict_thr  Decision threshold(s) applied to predict data, allowing predict to be
                      continuous probabilities or decisions
         truth_thr    Decision threshold(s) applied to truth data, allowing truth to be
                      continuous probabilities or decisions
         timesteps

       Default predict_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
       if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
       the confusion matrix is calculated for all classes.

       Returns pandas dataframe (snrdf) of metrics per SNR.
    """
    import warnings

    from sonusai.metrics import one_hot
    from sonusai.queries import get_mixids_from_snr

    num_classes = truth_f.shape[1]

    snr_mixids = get_mixids_from_snr(mixdb=mixdb, mixids=mixid)

    # Check predict_thr array or scalar and return final scalar predict_thr value
    if not mixdb.truth_mutex and num_classes > 1:
        if np.ndim(predict_thr) == 0 and predict_thr == 0:
            # multi-label predict_thr scalar 0 force to 0.5 default
            predict_thr = 0.5

        if np.ndim(predict_thr) == 1:
            if len(predict_thr) == 1:
                if predict_thr[0] == 0:
                    # multi-label predict_thr array scalar 0 force to 0.5 default
                    predict_thr = 0.5
                else:
                    # multi-label predict_thr array set to scalar = array[0]
                    predict_thr = predict_thr[0]

    macro_avg = np.zeros((len(snr_mixids), 7), dtype=np.float32)
    micro_avg = np.zeros((len(snr_mixids), 7), dtype=np.float32)
    wghtd_avg = np.zeros((len(snr_mixids), 7), dtype=np.float32)
    ssnr_stats = None
    segsnr_f = None

    if segsnr is not None:
        # prep segsnr if provided, transform frames to feature frames via mean()
        # expected to always be an integer
        feature_frames = int(segsnr.shape[0] / truth_f.shape[0])
        segsnr_f = np.mean(np.reshape(segsnr, (truth_f.shape[0], feature_frames)), axis=1, keepdims=True)
        ssnr_stats = np.zeros((len(snr_mixids), 3), dtype=np.float32)

    ii = 0
    for snr in snr_mixids:
        y_truth, y_predict = get_mixids_data(mixdb, snr_mixids[snr], truth_f, predict)
        _, metrics, _, _, _, mavg = one_hot(y_truth, y_predict, predict_thr, truth_thr, timesteps)

        # mavg macro, micro, weighted: [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
        macro_avg[ii, :] = mavg[0, 0:7]
        micro_avg[ii, :] = mavg[1, 0:7]
        wghtd_avg[ii, :] = mavg[2, 0:7]
        if segsnr is not None:
            y_truth, y_segsnr = get_mixids_data(mixdb, snr_mixids[snr], truth_f, segsnr_f)
            with warnings.catch_warnings():
                warnings.filterwarnings(action='ignore', message='divide by zero encountered in log10')
                # segmental SNR mean = mixture_snr and target_snr
                ssnr_stats[ii, 0] = 10 * np.log10(np.mean(y_segsnr))
                # segmental SNR 80% percentile
                ssnr_stats[ii, 1] = 10 * np.log10(np.percentile(y_segsnr, 80, method='midpoint'))
                # segmental SNR max
                ssnr_stats[ii, 2] = 10 * np.log10(max(y_segsnr))

        ii += 1

    # SNR format: PPV, TPR, F1, FPR, ACC, AP, AUC
    col_n = ['PPV', 'TPR', 'F1', 'FPR', 'ACC', 'AP', 'AUC']
    snr_macrodf = pd.DataFrame(macro_avg, index=list(snr_mixids.keys()), columns=col_n)
    snr_macrodf.sort_index(ascending=False, inplace=True)

    snr_microdf = pd.DataFrame(micro_avg, index=list(snr_mixids.keys()), columns=col_n)
    snr_microdf.sort_index(ascending=False, inplace=True)

    snr_wghtdf = pd.DataFrame(wghtd_avg, index=list(snr_mixids.keys()), columns=col_n)
    snr_wghtdf.sort_index(ascending=False, inplace=True)

    # Add segmental SNR columns if provided
    if segsnr is not None:
        ssnrdf = pd.DataFrame(ssnr_stats, index=list(snr_mixids.keys()), columns=['SSNRavg', 'SSNR80p', 'SSNRmax'])
        ssnrdf.sort_index(ascending=False, inplace=True)
        snr_macrodf = pd.concat([snr_macrodf, ssnrdf], axis=1)
        snr_microdf = pd.concat([snr_microdf, ssnrdf], axis=1)
        snr_wghtdf = pd.concat([snr_wghtdf, ssnrdf], axis=1)

    return snr_macrodf, snr_microdf, snr_wghtdf, snr_mixids



"""sonusai vars

usage: vars [-h]

options:
   -h, --help   Display this help.

List custom SonusAI variables.

"""


def main():
    from docopt import docopt

    import sonusai
    from sonusai.mixture import DEFAULT_NOISE
    from sonusai.utils import trim_docstring

    docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    print('Custom SonusAI variables:')
    print('')
    print(f'  ${{default_noise}}: {DEFAULT_NOISE}')
    print('')


if __name__ == '__main__':
    main()



"""sonusai lsdb

usage: lsdb [-hta] [-i MIXID] [-c TID] LOC

Options:
    -h, --help
    -i MIXID, --mixid MIXID         Mixture ID(s) to analyze. [default: *].
    -c TID, --truth_index TID       Analyze mixtures that contain this truth index.
    -t, --targets                   List all target files.
    -a, --all_class_counts          List all class counts.

List mixture data information from a SonusAI mixture database.

Inputs:
    LOC     A SonusAI mixture database directory.

"""
import numpy as np

from sonusai import logger
from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import MixtureDatabase


def lsdb(mixdb: MixtureDatabase,
         mixids: GeneralizedIDs = None,
         truth_index: int = None,
         list_targets: bool = False,
         all_class_counts: bool = False) -> None:
    import h5py

    from sonusai import SonusAIError
    from sonusai.mixture import calculate_snr_f_statistics
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import get_truth_indices_for_target
    from sonusai.queries import get_mixids_from_truth_index
    from sonusai.utils import consolidate_range
    from sonusai.utils import print_mixture_details
    from sonusai.utils import seconds_to_hms

    desc_len = 24

    total_samples = mixdb.total_samples()
    total_duration = total_samples / SAMPLE_RATE

    logger.info(f'{"Mixtures":{desc_len}} {len(mixdb.mixtures)}')
    logger.info(f'{"Duration":{desc_len}} {seconds_to_hms(seconds=total_duration)}')
    logger.info(f'{"Targets":{desc_len}} {len(mixdb.targets)}')
    logger.info(f'{"Target augmentations":{desc_len}} {len(mixdb.target_augmentations)}')
    logger.info(f'{"Noise augmentations":{desc_len}} {len(mixdb.noise_augmentations)}')
    logger.info(f'{"Noises":{desc_len}} {len(mixdb.noises)}')
    logger.info(f'{"Feature":{desc_len}} {mixdb.feature}')
    logger.info(
        f'{"Feature shape":{desc_len}} {mixdb.fg.stride} x {mixdb.fg.num_bands} ({mixdb.fg.stride * mixdb.fg.num_bands} total params)')
    logger.info(f'{"Feature samples":{desc_len}} {mixdb.feature_samples} samples ({mixdb.feature_ms} ms)')
    logger.info(
        f'{"Feature step samples":{desc_len}} {mixdb.feature_step_samples} samples ({mixdb.feature_step_ms} ms)')
    logger.info(f'{"Feature overlap":{desc_len}} {mixdb.fg.step / mixdb.fg.stride} ({mixdb.feature_step_ms} ms)')
    logger.info(f'{"SNRs":{desc_len}} {mixdb.snrs}')
    logger.info(f'{"Random SNRs":{desc_len}} {mixdb.random_snrs}')
    logger.info(f'{"Classes":{desc_len}} {mixdb.num_classes}')
    logger.info(f'{"Truth mutex":{desc_len}} {mixdb.truth_mutex}')
    # TODO: fix class count
    logger.info(f'{"Class count":{desc_len}} not supported')
    # print_class_count(class_count=class_count, length=desc_len, print_fn=logger.info)
    # TODO: add class weight calculations here
    logger.info('')

    if list_targets:
        logger.info('Target details:')
        idx_len = int(np.ceil(np.log10(len(mixdb.targets))))
        for idx, target in enumerate(mixdb.targets):
            desc = f'  {idx:{idx_len}} Name'
            logger.info(f'{desc:{desc_len}} {target.name}')
            desc = f'  {idx:{idx_len}} Truth index'
            logger.info(f'{desc:{desc_len}} {get_truth_indices_for_target(target)}')
        logger.info('')

    if truth_index is not None:
        if 0 <= truth_index > mixdb.num_classes:
            raise SonusAIError(f'Given truth_index is outside valid range of 1-{mixdb.num_classes}')
        ids = get_mixids_from_truth_index(mixdb=mixdb, predicate=lambda x: x in [truth_index])[truth_index]
        logger.info(f'Mixtures with truth index {truth_index}: {ids}')
        logger.info('')

    mixids = mixdb.mixids_to_list(mixids)

    if len(mixids) == 1:
        print_mixture_details(mixdb=mixdb, mixid=mixids[0], desc_len=desc_len, print_fn=logger.info)
        if all_class_counts:
            # TODO: fix class count
            logger.info('All class count not supported')
            # print_class_count(class_count=class_count, length=desc_len, print_fn=logger.info, all_class_counts=True)
    else:
        logger.info(f'Calculating statistics from truth_f files for {len(mixids):,} mixtures'
                    f' ({consolidate_range(mixids)})')
        for mixid in mixids:
            with h5py.File(mixdb.mixture_filename(mixid), 'r') as f:
                if mixid == mixids[0]:
                    truth_f = np.array(f['truth_f'])
                else:
                    truth_f = np.concatenate((truth_f, np.array(f['truth_f'])))

        snr_mean, snr_std, snr_db_mean, snr_db_std = calculate_snr_f_statistics(truth_f)

        logger.info('Truth')
        logger.info(f'  {"mean":^8s}  {"std":^8s}  {"db_mean":^8s}  {"db_std":^8s}')
        for c in range(len(snr_mean)):
            logger.info(f'  {snr_mean[c]:8.2f}  {snr_std[c]:8.2f}  {snr_db_mean[c]:8.2f}  {snr_db_std[c]:8.2f}')


def main():
    from docopt import docopt

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    mixid = args['--mixid']
    truth_index = args['--truth_index']
    list_targets = args['--targets']
    all_class_counts = args['--all_class_counts']
    location = args['LOC']

    if truth_index is not None:
        truth_index = int(truth_index)

    create_file_handler('lsdb.log')
    update_console_handler(False)
    initial_log_messages('lsdb')

    logger.info(f'Analyzing {location}')

    mixdb = MixtureDatabase(config=location, show_progress=True)
    lsdb(mixdb=mixdb,
         mixids=mixid,
         truth_index=truth_index,
         list_targets=list_targets,
         all_class_counts=all_class_counts)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from typing import Callable

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import ClassCount


def print_mixture_details(mixdb: MixtureDatabase,
                          mixid: int = None,
                          desc_len: int = 1,
                          print_fn: Callable = print) -> None:
    import numpy as np

    from sonusai import SonusAIError
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import seconds_to_hms

    if mixid is not None:
        if 0 < mixid >= len(mixdb.mixtures):
            raise SonusAIError(f'Given mixid is outside valid range of 0:{len(mixdb.mixtures) - 1}.')

        print_fn(f'Mixture {mixid} details')
        mixture = mixdb.mixtures[mixid]
        targets = [mixdb.targets[idx] for idx in mixture.target_file_index]
        target_augmentations = [mixdb.target_augmentations[idx] for idx in mixture.target_augmentation_index]
        noise = mixdb.noises[mixture.noise_file_index]
        for t_idx, target in enumerate(targets):
            print_fn(f'  Target {t_idx}')
            print_fn(f'{"    Name":{desc_len}} {target.name}')
            print_fn(f'{"    Duration":{desc_len}} {seconds_to_hms(target.duration)}')
            for ts_idx, truth_setting in enumerate(target.truth_settings):
                print_fn(f'    Truth setting {ts_idx}')
                print_fn(f'{"      Index":{desc_len}} {truth_setting.index}')
                print_fn(f'{"      Function":{desc_len}} {truth_setting.function}')
                print_fn(f'{"      Config":{desc_len}} {truth_setting.config}')
            print_fn(f'{"    Augmentation":{desc_len}} {target_augmentations[t_idx]}')
        print_fn(f'{"  Samples":{desc_len}} {mixture.samples}')
        print_fn(f'{"  Feature frames":{desc_len}} {mixdb.mixture_feature_frames(mixid)}')
        print_fn(f'{"  Noise file":{desc_len}} {noise.name}')
        noise_offset_percent = int(np.round(100 * mixture.noise_offset / float(noise.duration * SAMPLE_RATE)))
        print_fn(f'{"  Noise offset":{desc_len}} {mixture.noise_offset} samples ({noise_offset_percent}%)')
        print_fn(f'{"  SNR":{desc_len}} {mixture.snr} dB{" (random)" if mixture.random_snr else ""}')
        print_fn(f'{"  Target gain":{desc_len}} {mixture.target_gain}')
        print_fn(f'{"  Target SNR gain":{desc_len}} {mixture.target_snr_gain}')
        print_fn(f'{"  Noise SNR gain":{desc_len}} {mixture.noise_snr_gain}')
        print_fn('')


def print_class_count(class_count: ClassCount,
                      length: int,
                      print_fn: Callable = print,
                      all_class_counts: bool = False) -> None:
    import numpy as np

    print_fn(f'Class count:')
    idx_len = int(np.ceil(np.log10(len(class_count))))
    for idx, count in enumerate(class_count):
        if all_class_counts or count > 0:
            desc = f'  class {idx + 1:{idx_len}}'
            print_fn(f'{desc:{length}} {count}')



from typing import List

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import AugmentedTarget
from sonusai.mixture.types import AugmentedTargets
from sonusai.mixture.types import TargetFile
from sonusai.mixture.types import TargetFiles


def get_augmented_targets(target_files: TargetFiles, target_augmentations: Augmentations) -> AugmentedTargets:
    from sonusai.mixture import get_augmentation_indices_for_mixup
    from sonusai.mixture import get_mixups

    mixups = get_mixups(target_augmentations)

    augmented_targets: AugmentedTargets = []
    for mixup in mixups:
        augmentation_indices = get_augmentation_indices_for_mixup(target_augmentations, mixup)
        for target_index in range(len(target_files)):
            for augmentation_index in augmentation_indices:
                augmented_targets.append(AugmentedTarget(target_file_index=target_index,
                                                         target_augmentation_index=augmentation_index))

    return augmented_targets


def get_truth_indices_for_target(target: TargetFile) -> List[int]:
    """Get a list of truth indices for a given target."""
    index = [truth_setting.index for truth_setting in target.truth_settings]

    # flatten, uniquify, and sort
    return sorted(list(set([item for sublist in index for item in sublist])))


def get_truth_indices_for_augmented_target(augmented_target: AugmentedTarget, targets: TargetFiles) -> List[int]:
    return get_truth_indices_for_target(targets[augmented_target.target_file_index])


def get_mixup_for_augmented_target(augmented_target: AugmentedTarget, augmentations: Augmentations) -> int:
    return augmentations[augmented_target.target_augmentation_index].mixup


def get_target_indices_for_truth_index(targets: TargetFiles,
                                       truth_index: int,
                                       allow_multiple: bool = False) -> List[int]:
    """Get a list of target indices containing the given truth index.

    If allow_multiple is True, then include targets that contain multiple truth indices.
    """
    target_indices = set()
    for target_index, target in enumerate(targets):
        indices = get_truth_indices_for_target(target)
        if len(indices) == 1 or allow_multiple:
            for index in indices:
                if index == truth_index + 1:
                    target_indices.add(target_index)

    return sorted(list(target_indices))


def get_augmented_target_indices_for_truth_index(augmented_targets: AugmentedTargets,
                                                 targets: TargetFiles,
                                                 augmentations: Augmentations,
                                                 truth_index: int,
                                                 mixup: int,
                                                 allow_multiple: bool = False) -> List[int]:
    """Get a list of augmented target indices containing the given truth index.

    If allow_multiple is True, then include targets that contain multiple truth indices.
    """
    augmented_target_indices = set()
    for augmented_target_index, augmented_target in enumerate(augmented_targets):
        if get_mixup_for_augmented_target(augmented_target=augmented_target, augmentations=augmentations) == mixup:
            indices = get_truth_indices_for_augmented_target(augmented_target=augmented_target, targets=targets)
            if len(indices) == 1 or allow_multiple:
                for index in indices:
                    if index == truth_index + 1:
                        augmented_target_indices.add(augmented_target_index)

    return sorted(list(augmented_target_indices))


def get_augmented_target_indices_by_class(mixdb: MixtureDatabase,
                                          augmented_targets: AugmentedTargets,
                                          mixup: int) -> List[List[int]]:
    num_classes = mixdb.num_classes
    if mixdb.truth_mutex:
        num_classes -= 1

    indices = []
    for idx in range(num_classes):
        indices.append(
            get_augmented_target_indices_for_truth_index(augmented_targets=augmented_targets,
                                                         targets=mixdb.targets,
                                                         augmentations=mixdb.target_augmentations,
                                                         truth_index=idx,
                                                         mixup=mixup))
    return indices


def get_augmented_target_indices_for_mixup(mixdb: MixtureDatabase,
                                           augmented_targets: AugmentedTargets,
                                           mixup: int) -> List[List[int]]:
    from collections import deque
    from random import shuffle

    from sonusai import SonusAIError

    mixup_indices = []

    if mixup == 1:
        for index, augmented_target in enumerate(augmented_targets):
            if get_mixup_for_augmented_target(augmented_target=augmented_target,
                                              augmentations=mixdb.target_augmentations) == 1:
                mixup_indices.append([index])
        return mixup_indices

    augmented_target_indices_by_class = get_augmented_target_indices_by_class(mixdb=mixdb,
                                                                              augmented_targets=augmented_targets,
                                                                              mixup=mixup)

    if mixup > mixdb.num_classes:
        raise SonusAIError(
            f'Specified mixup, {mixup}, is greater than the number of classes, {mixdb.num_classes}')

    de = deque()

    # Keep looping until not enough targets remain for mixup
    while sum([1 for x in augmented_target_indices_by_class if x]) >= mixup:
        # Need more class indices?
        if len(de) < mixup:
            # Only choose classes that still have data
            counts = [len(item) for item in augmented_target_indices_by_class]
            # Need to subtract out indices already in the deque
            for idx in de:
                counts[idx] -= 1
            indices = [idx for idx, val in enumerate(counts) if val > 0]
            shuffle(indices)
            # Keep shuffling if the deque is not empty and the first new index matches the last item
            # (so that a class does not appear twice in a mixup)
            while de and indices[0] == de[-1]:
                shuffle(indices)
            for index in indices:
                de.append(index)

        class_indices = [de.popleft() for _ in range(mixup)]

        target_indices = []
        for class_index in class_indices:
            target_indices.append(augmented_target_indices_by_class[class_index].pop())

        mixup_indices.append(target_indices)

    return mixup_indices



def create_ts_name(base: str) -> str:
    """Create a timestamped name based on 'base'."""
    from datetime import datetime
    from os.path import exists

    ts = datetime.now()

    # First try just date
    dir_name = base + '-' + ts.strftime('%Y%m%d')
    if exists(dir_name):
        # add hour-min-sec if necessary
        dir_name = base + '-' + ts.strftime('%Y%m%d-%H%M%S')

    return dir_name



import numpy as np

from sonusai.mixture.types import Predict


def read_predict_data(filename: str) -> Predict:
    """Read predict data from given HDF5 file and return it."""
    import h5py

    from sonusai import SonusAIError
    from sonusai import logger

    logger.debug(f'Reading prediction data from {filename}')
    with h5py.File(filename, 'r') as f:
        # prediction data is either [frames, num_classes], or [frames, timesteps, num_classes]
        predict = np.array(f['predict'])

        if predict.ndim == 2:
            return predict

        if predict.ndim == 3:
            frames, timesteps, num_classes = predict.shape

            logger.debug(f'Reshaping prediction data in {filename} 'f''
                         f'from [{frames}, {timesteps}, {num_classes}] '
                         f'to [{frames * timesteps}, {num_classes}]')
            predict = np.reshape(predict, [frames * timesteps, num_classes], order='F')
            return predict

        raise SonusAIError(f'Invalid prediction data dimensions in {filename}')



from pyaaware import ForwardTransform
from pyaaware import InverseTransform

from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import ImpulseResponseData
from sonusai.mixture.types import Location


def get_next_noise(audio: AudioT, offset: int, length: int) -> AudioT:
    """Get next sequence of noise data from noise audio file

    :param audio: Overall noise audio (entire file's worth of data)
    :param offset: Starting sample
    :param length: Number of samples to get
    :return: Sequence of noise audio data
    """
    import numpy as np

    return np.take(audio, range(offset, offset + length), mode='wrap')


def read_ir(name: Location) -> ImpulseResponseData:
    """Read impulse response data from a file

    :param name: File name
    :return: ImpulseResponseData object
    """
    import numpy as np
    from scipy.io import wavfile

    from sonusai import SonusAIError
    from sonusai.mixture import ImpulseResponseData
    from sonusai.mixture import apply_ir
    from sonusai.mixture import tokenized_expandvars

    expanded_name, _ = tokenized_expandvars(name)

    try:
        # Read in and normalize to -20 dBFS to avoid clipping when applying IR
        sample_rate, data = wavfile.read(expanded_name)
        max_data = max(abs(data)) * 10
        data = list(np.array(data / max_data, dtype=np.float32))

        # Find offset to align convolved audio with original
        ir = ImpulseResponseData(sample_rate=sample_rate, offset=0, filter=data)
        x = np.zeros((len(ir.filter),), dtype=np.float32)
        x[0] = 1
        y = list(apply_ir(x, ir))

        return ImpulseResponseData(sample_rate=sample_rate, offset=y.index(max(y)), filter=data)
    except Exception as e:
        if name != expanded_name:
            raise SonusAIError(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')


def read_audio(name: Location) -> AudioT:
    """Read audio data from a file

    :param name: File name
    :return: Array of time domain audio data
    """
    import numpy as np
    import sox

    from sonusai import SonusAIError
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import ENCODING
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import tokenized_expandvars

    expanded_name, _ = tokenized_expandvars(name)

    if BIT_DEPTH == 8:
        encoding_out = np.int8
    elif BIT_DEPTH == 16:
        encoding_out = np.int16
    elif BIT_DEPTH == 24:
        encoding_out = np.int32
    elif BIT_DEPTH == 32:
        if ENCODING == 'floating-point':
            encoding_out = np.float32
        else:
            encoding_out = np.int32
    elif BIT_DEPTH == 64:
        encoding_out = np.float64
    else:
        raise SonusAIError(f'Invalid BIT_DEPTH {BIT_DEPTH}')

    try:
        # Read in and convert to desired format
        # NOTE: pysox format transformations do not handle encoding properly; need to use direct call to sox instead
        args = ['-D',
                '-V2',
                '-c', '1',
                expanded_name,
                '-t', 'raw',
                '-r', str(SAMPLE_RATE),
                '-b', str(BIT_DEPTH),
                '-c', str(CHANNEL_COUNT),
                '-e', ENCODING,
                '-']
        status, out, err = sox.core.sox(args, None, False)
        if status != 0:
            raise SonusAIError(f'sox stdout: {out}\nsox stderr: {err}')

        return np.frombuffer(out, dtype=encoding_out)

    except Exception as e:
        if name != expanded_name:
            raise SonusAIError(f'Error reading {name} (expanded: {expanded_name}):\n{e}')
        else:
            raise SonusAIError(f'Error reading {name}:\n{e}')


def calculate_transform_from_audio(audio: AudioT, transform: ForwardTransform) -> AudioF:
    """Apply forward transform to input audio data to generate transform data

    :param audio: Time domain data [samples]
    :param transform: ForwardTransform object
    :return: Frequency domain data [frames, bins]
    """
    return transform.execute_all(audio).transpose()


def calculate_audio_from_transform(data: AudioF, transform: InverseTransform, trim: bool = True) -> AudioT:
    """Apply inverse transform to input transform data to generate audio data

    :param data: Frequency domain data [frames, bins]
    :param transform: InverseTransform object
    :param trim: Removes starting samples so output waveform will be time-aligned with input waveform to the transform
    :return: Time domain data [samples]
    """
    audio = transform.execute_all(data.transpose())
    if trim:
        audio = audio[transform.N - transform.R:]

    return audio



"""sonusai calc_metric_spenh_targetf

usage: calc_metric_spenh_targetf [-hvtpws] [-i MIXID] (-d PLOC) INPUT

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to generate. [default: *].
    -d PLOC, --ploc PLOC        Location of SonusAI predict data.
    -t, --truth-est-mode        Calculate extraction using truth and include metrics.
    -p, --plot                  Enable PDF plots file generation per mixture.
    -w, --wav                   Enable WAV file generation per mixture.
    -s, --summary               Enable summary files generation.

Calculate speech enhancement target_f metrics for prediction data in PLOC and SonusAI mixture database in INPUT.

Inputs:
    PLOC    SonusAI prediction data directory.
    INPUT   SonusAI mixture database directory.

Outputs the following to PLOC:
    <id>_metric_spenh_targetf.txt

    If --plot:
        <id>_metric_spenh_targetf.pdf

    If --wav:
        <id>_target.wav
        <id>_target_est.wav
        <id>_noise.wav
        <id>_noise_est.wav
        <id>_mixture.wav

        If --truth-est-mode:
            <id>_target_truth_est.wav
            <id>_noise_truth_est.wav

    If --summary:
        metric_spenh_targetf_summary.txt
        metric_spenh_targetf_summary.csv
        metric_spenh_targetf_list.csv
        metric_spenh_targetf_estats_list.csv

        If --truth-est-mode:
            metric_spenh_targetf_truth_list.csv
            metric_spenh_targetf_estats_truth_list.csv

"""
from dataclasses import dataclass
from typing import Tuple
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from sonusai import logger
from sonusai.mixture import AudioF
from sonusai.mixture import AudioT
from sonusai.mixture import Feature
from sonusai.mixture import Location
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import Predict


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    predict_location: Location = None
    truth_est_mode: bool = None
    enable_plot: bool = None
    enable_wav: bool = None


MP_GLOBAL = MPGlobal()


def calc_pcm(s_true: np.ndarray, s_est: np.ndarray, with_log: bool = False) -> np.ndarray:
    """Calculate pcm ((Phase Constrained Magnitude)) err/oss using all source inputs
    which are frames x nsrc x bins. These should include a noise to
    be a complete mixture estimate, i.e. noise_est = mixture - sum-over-n(s_est(:,n,:))
    should be one of the sources in s_true and s_est.
    Calculates mean-over-srcs(mean-over-tf(| (|Sr(t,f)|+|Si(t,f)|) - (|Shr(t,f)|+|Shi(t,f)|) |))
    log is optional if with_log=True (for use as a loss function)
    Returns:
      pcm scalar
    ref:  2022-04-sa-sdr-novel-loss-for-separation
    """

    # multisrc sa-sdr, inputs must be timesteps x nsrc
    err = s_true - s_est  # samples x nsrc

    # -10*log10(sumk(||sk||^2) / sumk(||sk - shk||^2)
    num = np.sum(s_true ** 2)  # sum over samples and sources
    den = np.sum(err ** 2)
    if num == 0 and den == 0:
        ratio = np.inf
    else:
        ratio = num / (den + np.finfo(np.float32).eps)

    sa_sdr = 10 * np.log10(ratio)

    return sa_sdr


def calc_sa_sdr(s_true: np.ndarray,
                s_est: np.ndarray,
                with_scale: bool = False,
                with_negate: bool = False) -> (np.ndarray, np.ndarray):
    """Calculate sa-sdr (source-aggregated SDR (signal distortion ratio)) using all source
    inputs which are samples,nsrc time-domain inputs. These should include a noise to
    be a complete mixture estimate, i.e. noise_est = mixture - sum-over-n(s_est(:,n,:))
    should be one of the sources in s_true and s_est.
    Calculates -10*log10(sumn(||sn||^2) / sumn(||sn - shn||^2)
    Scaling is optional if with_scale=True, scaling is same as in SI-SDR
    Negation is optional if with_negate=True (for use as a loss function)
    Returns:
      sa_sdr scalar
    ref:  2022-04-sa-sdr-novel-loss-for-separation
    """

    if with_scale:
        # calc 1 x nsrc scaling factors
        ref_energy = np.sum(s_true ** 2, axis=0, keepdims=True)  # 1,nsrc
        # if ref_energy is zero, just set scaling to 1.0  # 1,nsrc
        with np.errstate(divide='ignore', invalid='ignore'):
            opt_scale = np.sum(s_true * s_est, axis=0, keepdims=True) / ref_energy
            opt_scale[opt_scale == np.inf] = 1.0
            opt_scale = np.nan_to_num(opt_scale, nan=1.0)
        scaled_ref = opt_scale * s_true  # 1 x nsrc
    else:
        scaled_ref = s_true
        opt_scale = np.ones((1, s_true.shape[1]), dtype=float)

    # multisrc sa-sdr, inputs must be timesteps x nsrc
    err = scaled_ref - s_est  # samples x nsrc

    # -10*log10(sumk(||sk||^2) / sumk(||sk - shk||^2)
    num = np.sum(s_true ** 2)  # sum over samples and sources
    den = np.sum(err ** 2)
    if num == 0 and den == 0:
        ratio = np.inf
    else:
        ratio = num / (den + np.finfo(np.float32).eps)

    sa_sdr = 10 * np.log10(ratio)

    if with_negate:
        sa_sdr = -sa_sdr  # for use as a loss function

    return sa_sdr, opt_scale


def mean_square_error(ytrue: np.ndarray,
                      ypred: np.ndarray,
                      squared: bool = False) -> (np.ndarray, np.ndarray, np.ndarray):
    """Calculate mean square error or root mean square error between ytrue, ypred

    Typical inputs expected to be size frames x classes/bins
    returns:
      err    mean over both dimensions
      err_c  mean-over-dim0 i.e. a value per class/bin
      err_f  mean-over-dim1 i.e. a value per frame
    """
    sqerr = np.square(ytrue - ypred)
    err_c = np.mean(sqerr, axis=0)  # mean over frames for value per class
    err_f = np.mean(sqerr, axis=1)  # mean over classes for value per frame
    err = np.mean(sqerr)  # mean over all
    if not squared:
        err_c = np.sqrt(err_c)
        err_f = np.sqrt(err_f)
        err = np.sqrt(err)

    return err, err_c, err_f


def mean_abs_percentage_error(ytrue: np.ndarray, ypred: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
    """Calculate mean abs percentage error between ytrue, ypred
    If inputs are complex, calculates average:  mape(real)/2 + mape(imag)/2

    Inputs expected to be size frames x classes/bins
    returns:
      err    mean over both dimensions
      err_c  mean-over-dim0 i.e. a value per class/bin
      err_f  mean-over-dim1 i.e. a value per frame
    """
    if np.iscomplexobj(ytrue) is not True and np.iscomplexobj(ypred) is not True:
        abserr = 100 * np.abs((ytrue - ypred) / (ytrue + np.finfo(np.float32).eps))
    else:
        ytrue_r = np.real(ytrue)
        ytrue_i = np.imag(ytrue)
        ypred_r = np.real(ypred)
        ypred_i = np.imag(ypred)
        abserr_r = 100 * np.abs((ytrue_r - ypred_r) / (ytrue_r + np.finfo(np.float32).eps))
        abserr_i = 100 * np.abs((ytrue_i - ypred_i) / (ytrue_i + np.finfo(np.float32).eps))
        abserr = (abserr_r / 2) + (abserr_i / 2)

    err_c = np.mean(abserr, axis=0)  # mean over frames for value per class
    err_f = np.mean(abserr, axis=1)  # mean over classes for value per frame
    err = np.mean(abserr)  # mean over all

    return err, err_c, err_f


def log_error(ytrue: np.ndarray, ypred: np.ndarray) -> (np.ndarray, np.ndarray, np.ndarray):
    """Calculate log error between ytrue, ypred which are complex or real float numpy arrays

    Typical inputs expected to be size frames x classes/bins
    returns:
      logerr    mean over both dimensions, scalar value
      logerr_c  mean-over-dim0 i.e. a value per class/bin
      logerr_f  mean-over-dim1 i.e. a value per frame
    """
    ytsq = np.real(ytrue * np.conjugate(ytrue))
    ypsq = np.real(ypred * np.conjugate(ypred))
    lgerr = abs(10 * np.log10((ytsq + np.finfo(np.float32).eps) / (ypsq + np.finfo(np.float32).eps)))
    # lgerr = abs(10*np.log10(ytsq / (ypsq + np.finfo(np.float32).eps)+np.finfo(np.float32).eps))
    logerr_c = np.mean(lgerr, axis=0)  # mean over frames for value per class
    logerr_f = np.mean(lgerr, axis=1)  # mean over classes for value per frame
    logerr = np.mean(lgerr)

    return logerr, logerr_c, logerr_f


def calculate_pesq(speech_ref: np.ndarray, speech_est: np.ndarray, error_value=0.0):
    """Computes the PESQ score of speech estimate audio vs. the clean speech estimate audio

    Upon error, assigns a value of 0, or user specified value in error_value
    Argument/s:
      speech_ref - speech reference audio, np.ndarray vector of samples.
      speech_est - speech estimated audio, np.ndarray vector of samples.
    Returns:
      score - float value between -0.5 to 4.5.
   """
    import warnings

    from pesq import pesq

    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            score = pesq(16000, speech_ref, speech_est, mode='wb')
    except Exception as e:
        logger.debug(f'PESQ error {e}')
        score = error_value

    return score


def plot_mixpred(mixture: AudioT,
                 mixture_f: AudioF,
                 target: AudioT = None,
                 feature: Feature = None,
                 predict: Predict = None,
                 tp_title: str = '') -> plt.figure:
    from sonusai.mixture import SAMPLE_RATE

    num_plots = 2
    if feature is not None:
        num_plots += 1
    if predict is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    p = 0
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[p].plot(x_axis, mixture, label='Mixture', color='mistyrose')
    ax[0].set_ylabel('magnitude', color='tab:blue')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    # ax[p].set_ylim([-1.025, 1.025])
    if target is not None:  # Plot target time-domain waveform on top of mixture
        ax[0].plot(x_axis, target, label='Target', color='tab:blue')
        # ax[0].tick_params(axis='y', labelcolor=color)
    ax[p].set_title('Waveform')

    # Plot the mixture spectrogram
    p += 1
    ax[p].imshow(np.transpose(mixture_f), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Mixture')

    if feature is not None:
        p += 1
        ax[p].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Feature')

    if predict is not None:
        p += 1
        ax[p].imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Predict ' + tp_title)

    return fig


def plot_pdb_predtruth(predict: np.ndarray,
                       truth_f: Union[np.ndarray, None] = None,
                       metric: Union[np.ndarray, None] = None,
                       tp_title: str = '') -> plt.figure:
    """Plot predict and optionally truth and a metric in power db, e.g. applies 10*log10(predict)"""

    num_plots = 2
    if truth_f is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    tmp = 10 * np.log10(predict.transpose() + np.finfo(np.float32).eps)
    ax[p].imshow(tmp, aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Predict')

    if truth_f is not None:
        p += 1
        tmp = 10 * np.log10(truth_f.transpose() + np.finfo(np.float32).eps)
        ax[p].imshow(tmp, aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    # Plot the predict avg, and optionally truth avg and metric lines
    pred_avg = 10 * np.log10(np.mean(predict, axis=-1) + np.finfo(np.float32).eps)
    p += 1
    x_axis = np.arange(len(pred_avg), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, pred_avg, color='black', linestyle='dashed', label='Predict mean over freq.')
    ax[p].set_ylabel('mean db', color='black')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_f is not None:
        truth_avg = 10 * np.log10(np.mean(truth_f, axis=-1) + np.finfo(np.float32).eps)
        ax[p].plot(x_axis, truth_avg, color='green', linestyle='dashed', label='Truth mean over freq.')

    if metric is not None:  # instantiate 2nd y-axis that shares the same x-axis
        ax2 = ax[p].twinx()
        color2 = 'red'
        ax2.plot(x_axis, metric, color=color2, label='sig distortion (mse db)')
        ax2.set_xlim(x_axis[0], x_axis[-1])
        ax2.set_ylim([0, np.max(metric)])
        ax2.set_ylabel('spectral distortion (mse db)', color=color2)
        ax2.tick_params(axis='y', labelcolor=color2)
        ax[p].set_title('SNR and SNR mse (mean over freq. db)')
    else:
        ax[p].set_title('SNR (mean over freq. db)')
        # ax[0].tick_params(axis='y', labelcolor=color)
    return fig


def plot_epredtruth(predict: np.ndarray,
                    predict_wav: np.ndarray,
                    truth_f: Union[np.ndarray, None] = None,
                    truth_wav: Union[np.ndarray, None] = None,
                    metric: Union[np.ndarray, None] = None,
                    tp_title: str = '') -> plt.figure:
    """Plot predict spectrogram and waveform and optionally truth and a metric)"""

    num_plots = 2
    if truth_f is not None:
        num_plots += 1
    if metric is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the predict spectrogram
    p = 0
    ax[p].imshow(predict.transpose(), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Predict')

    if truth_f is not None:
        p += 1
        ax[p].imshow(truth_f.transpose(), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth')

    # Plot the predict wav, and optionally truth avg and metric lines
    p += 1
    x_axis = np.arange(len(predict_wav), dtype=np.float32)  # / SAMPLE_RATE
    ax[p].plot(x_axis, predict_wav, color='black', linestyle='dashed', label='Speech Estimate')
    ax[p].set_ylabel('Amplitude', color='black')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    if truth_wav is not None:
        ntrim = len(truth_wav) - len(predict_wav)
        if ntrim > 0:
            truth_wav = truth_wav[0:-ntrim]
        ax[p].plot(x_axis, truth_wav, color='green', linestyle='dashed', label='True Target')

    # Plot the metric lines
    if metric is not None:
        p += 1
        x_axis = np.arange(len(metric), dtype=np.float32)  # / SAMPLE_RATE
        ax[p].plot(x_axis, metric, color='red', label='Target LogErr')
        ax[p].set_ylabel('log error db', color='red')
        ax[p].set_xlim(x_axis[0], x_axis[-1])
        ax[p].set_ylim([-0.01, np.max(metric) + .01])

    return fig


def _process_mixture(mixid: int) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    from os.path import basename
    from os.path import join
    from os.path import splitext

    import h5py

    from sonusai import SonusAIError
    from sonusai.utils import float_to_int16
    from sonusai.utils import unstack_complex
    from sonusai.utils import write_wav

    mixdb = MP_GLOBAL.mixdb
    predict_location = MP_GLOBAL.predict_location
    truth_est_mode = MP_GLOBAL.truth_est_mode
    enable_plot = MP_GLOBAL.enable_plot
    enable_wav = MP_GLOBAL.enable_wav

    # 1) Collect true target, noise, mixture data
    target = mixdb.mixture_target(mixid)
    target_f = mixdb.mixture_target_f(mixid, target=target)

    noise = mixdb.mixture_noise(mixid)
    noise_f = mixdb.mixture_noise_f(mixid, noise=noise)

    mixture = mixdb.mixture_mixture(mixid, target=target, noise=noise)
    mixture_f = mixdb.mixture_mixture_f(mixid, mixture=mixture)

    feature, truth_f = mixdb.mixture_ft(mixid, mixture=mixture)

    targetfi = mixdb.inverse_transform(target_f)  # need to use inv-tf to match size & tf properties
    noisefi = mixdb.inverse_transform(noise_f)  # need to use inv-tf to match size & tf properties

    # 2)  Read predict data
    output_name = join(predict_location, mixdb.mixtures[mixid].name)
    base_name = splitext(output_name)[0]
    try:
        with h5py.File(output_name, 'r') as f:
            predict = np.array(f['predict'])
    except Exception as e:
        raise SonusAIError(f'Error reading {output_name}: {e}')

    # 3) Extraction - for target_f truth, simply unstack as predict is estimating the target
    predict_complex = unstack_complex(predict)
    truth_f_complex = unstack_complex(truth_f)
    noise_est_complex = mixture_f - predict_complex
    target_est_wav = mixdb.inverse_transform(predict_complex)
    noise_est_wav = mixdb.inverse_transform(noise_est_complex)

    target_truth_est_complex = None
    target_truth_est_audio = None

    noise_truth_est_complex = None
    noise_truth_est_audio = None

    if truth_est_mode:
        # estimates using truth instead of prediction
        target_truth_est_complex = truth_f_complex
        target_truth_est_audio = mixdb.inverse_transform(target_truth_est_complex)

        noise_truth_est_complex = mixture_f - target_truth_est_complex
        noise_truth_est_audio = mixdb.inverse_transform(noise_truth_est_complex)

    # 4) Metrics
    # Mean absolute error (real and imag)
    cmape_tg, cmape_tg_bin, cmape_tg_frame = mean_abs_percentage_error(truth_f_complex, predict_complex)
    # Target/Speech logerr - PSD estimation accuracy symmetric mean log-spectral distortion
    lerr_tg, lerr_tg_bin, lerr_tg_frame = log_error(truth_f_complex, predict_complex)
    # Noise logerr - PSD estimation accuracy
    lerr_n, lerr_n_bin, lerr_n_frame = log_error(noise_f, noise_est_complex)
    # Noise td logerr
    # lerr_nt, lerr_nt_bin, lerr_nt_frame = log_error(noisefi, noise_truth_est_audio)

    # SA-SDR (time-domain source-aggragated SDR)
    ytrue = np.concatenate((targetfi[:, np.newaxis], noisefi[:, np.newaxis]), axis=1)
    ypred = np.concatenate((target_est_wav[:, np.newaxis], noise_est_wav[:, np.newaxis]), axis=1)
    sa_sdr, _ = calc_sa_sdr(ytrue, ypred)  # note: w/o scale is more pessimistic number

    # Speech intelligibility measure - PESQ
    pesq_speech = calculate_pesq(target, target_est_wav)
    pesq_mixture = calculate_pesq(target, mixture)
    pesq_impr = pesq_speech - pesq_mixture  # pesq improvement
    pesq_impr_pc = pesq_impr / (pesq_mixture + np.finfo(np.float32).eps) * 100  # pesq improvement %

    # 5) Save per mixture metric results
    # Single row in table of scalar metrics per mixture
    mtable1_col = ['MXSNR', 'MXPESQ', 'PESQ', 'PESQi%', 'SASDR', 'SPCMAPE%', 'SPLERR', 'NLERR', 'SPFILE', 'NFILE']
    ti = mixdb.mixtures[mixid].target_file_index[0]
    ni = mixdb.mixtures[mixid].noise_file_index
    metr1 = [mixdb.mixtures[mixid].snr, pesq_mixture, pesq_speech, pesq_impr_pc, sa_sdr, cmape_tg, lerr_tg,
             lerr_n, basename(mixdb.targets[ti].name), basename(mixdb.noises[ni].name)]
    mtab1 = pd.DataFrame([metr1], columns=mtable1_col, index=[mixid])

    # Stats of per frame estimation metrics
    efs_table2_col = ['Max', 'Min', 'Avg', 'Median']
    efs_table2_row = ['SPCMAPE%', 'SPLERR', 'NLERR']
    metr2 = [[np.max(cmape_tg_frame), np.min(cmape_tg_frame), np.mean(cmape_tg_frame), np.median(cmape_tg_frame)],
             [np.max(lerr_tg_frame), np.min(lerr_tg_frame), np.mean(lerr_tg_frame), np.median(lerr_tg_frame)],
             [np.max(lerr_n_frame), np.min(lerr_n_frame), np.mean(lerr_n_frame), np.median(lerr_n_frame)]]
    mtab2 = pd.DataFrame(metr2, columns=efs_table2_col, index=efs_table2_row)
    mtab2flat_col = ['MXSNR', 'SPCMAPE Max', 'SPCMAPE Min', 'SPCMAPE Avg', 'SPCMAPE Median',
                     'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
                     'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
    tmp = np.insert(np.array(metr2), 0, mixdb.mixtures[mixid].snr).reshape(1, 13)
    mtab2_flat = pd.DataFrame(tmp, columns=mtab2flat_col, index=[mixid])

    all_metrics_table_1 = mtab1
    all_metrics_table_2 = mtab2_flat

    metric_name = base_name + '_metric_spenh_targetf.txt'
    with open(metric_name, 'w') as f:
        print('Speech enhancement metrics:', file=f)
        print(mtab1.round(2).to_string(), file=f)
        print('', file=f)
        print(f'Extraction statistics over {mixture_f.shape[0]} frames:', file=f)
        print(mtab2.round(2).to_string(), file=f)
        print('', file=f)
        print(f'Target path: {mixdb.targets[ti].name}', file=f)
        print(f'Noise path: {mixdb.noises[ni].name}', file=f)
        # print(f'PESQ improvement: {pesq_impr:0.2f}, {pesq_impr_pc:0.1f}%', file=f)

    lerr_tgtr_frame = None
    lerr_ntr_frame = None
    all_metrics_table_3 = None
    all_metrics_table_4 = None

    if truth_est_mode:
        cmape_tgtr, cmape_tgtr_bin, cmape_tgtr_frame = mean_abs_percentage_error(truth_f_complex,
                                                                                 target_truth_est_complex)
        # metrics of estimates using truth instead of prediction
        lerr_tgtr, lerr_tgtr_bin, lerr_tgtr_frame = log_error(truth_f_complex, target_truth_est_complex)
        lerr_ntr, lerr_ntr_bin, lerr_ntr_frame = log_error(noise_f, noise_truth_est_complex)

        # ytrue = np.concatenate((targetfi[:, np.newaxis], noisefi[:, np.newaxis]), axis=1)
        ypred = np.concatenate((target_truth_est_audio[:, np.newaxis], noise_truth_est_audio[:, np.newaxis]), axis=1)
        sa_sdr_tr, opt_scale_tr = calc_sa_sdr(ytrue, ypred, with_scale=True)  # scale should be ones

        pesq_speechtr = calculate_pesq(target, target_truth_est_audio)
        pesq_impr_sptr = pesq_speechtr - pesq_mixture  # pesq improvement
        pesq_impr_pctr = pesq_impr_sptr / (pesq_mixture + np.finfo(np.float32).eps) * 100  # pesq improvement %

        mtable3_col = ['MXSNR', 'MXPESQ', 'PESQ', 'PESQi%', 'SASDR', 'SPCMAPE%', 'SPLERR', 'NLERR']
        metr3 = [mixdb.mixtures[mixid].snr, pesq_mixture, pesq_speechtr, pesq_impr_pctr, sa_sdr_tr,
                 cmape_tgtr, lerr_tgtr, lerr_ntr]
        mtab3 = pd.DataFrame([metr3], columns=mtable3_col, index=[mixid])

        # Stats of per frame estimation metrics
        efs_table4_col = ['Max', 'Min', 'Avg', 'Median']
        efs_table4_row = ['SPCMAPE%', 'SPLERR', 'NLERR']
        metr4 = [[np.max(cmape_tgtr_frame), np.min(cmape_tgtr_frame), np.mean(cmape_tgtr_frame),
                  np.median(cmape_tgtr_frame)],
                 [np.max(lerr_tgtr_frame), np.min(lerr_tgtr_frame), np.mean(lerr_tgtr_frame),
                  np.median(lerr_tgtr_frame)],
                 [np.max(lerr_ntr_frame), np.min(lerr_ntr_frame), np.mean(lerr_ntr_frame),
                  np.median(lerr_ntr_frame)]]
        mtab4 = pd.DataFrame(metr4, columns=efs_table4_col, index=efs_table4_row)

        # Append extraction metrics to metrics file:
        with open(metric_name, 'a') as f:
            print('', file=f)
            print('Speech enhancement metrics of extraction method using truth:', file=f)
            print(mtab3.round(2).to_string(), file=f)
            print('', file=f)
            print('Extraction (using Truth) statistics over frames:', file=f)
            print(mtab4.round(2).to_string(), file=f)

        # Append to all mixture table
        # mtab4flat_col = ['MXSNR', 'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
        #                  'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
        mtab4flat_col = ['MXSNR', 'SPCMAPE Max', 'SPCMAPE Min', 'SPCMAPE Avg', 'SPCMAPE Median',
                         'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
                         'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
        tmp = np.insert(np.array(metr4), 0, mixdb.mixtures[mixid].snr).reshape(1, 13)  # Insert MXSNR
        mtab4_flat = pd.DataFrame(tmp, columns=mtab4flat_col, index=[mixid])

        all_metrics_table_3 = mtab3
        all_metrics_table_4 = mtab4_flat

    # 7) write wav files
    if enable_wav:
        write_wav(name=base_name + '_mixture.wav', audio=float_to_int16(mixture))
        write_wav(name=base_name + '_target_est.wav', audio=float_to_int16(target_est_wav))
        write_wav(name=base_name + '_noise_est.wav', audio=float_to_int16(noise_est_wav))
        write_wav(name=base_name + '_target.wav', audio=float_to_int16(target))
        write_wav(name=base_name + '_noise.wav', audio=float_to_int16(noise))
        # debug code to test for perfect reconstruction of the extraction method
        # note both 75% olsa-hanns and 50% olsa-hann modes checked to have perfect reconstruction
        # target_r = mixdb.inverse_transform(target_f)
        # noise_r = mixdb.inverse_transform(noise_f)
        # _write_wav(name=base_name + '_target_r.wav', audio=float_to_int16(target_r))
        # _write_wav(name=base_name + '_noise_r.wav', audio=float_to_int16(noise_r)) # chk perfect rec
        if truth_est_mode:
            write_wav(name=base_name + '_target_truth_est.wav', audio=float_to_int16(target_truth_est_audio))
            write_wav(name=base_name + '_noise_truth_est.wav', audio=float_to_int16(noise_truth_est_audio))

    # 8) Write out plot file
    if enable_plot:
        from matplotlib.backends.backend_pdf import PdfPages
        plot_fname = base_name + '_metric_spenh_targetf.pdf'

        # Reshape feature to eliminate overlap redundancy for easier to understand spectrogram view
        # Original size (frames, stride, num_bands), decimates in stride dimension only if step is > 1
        # Reshape to get frames*decimated_stride, num_bands
        step = int(mixdb.feature_samples / mixdb.feature_step_samples)
        if feature.ndim != 3:
            raise SonusAIError(f'feature does not have 3 dimensions: frames, stride, num_bands')

        # for feature cn*00n**
        feat_sgram = unstack_complex(feature)
        feat_sgram = 20 * np.log10(abs(feat_sgram) + np.finfo(np.float32).eps)
        feat_sgram = feat_sgram[:, -step:, :]  # decimate,  Fx1xB
        feat_sgram = np.reshape(feat_sgram, (feat_sgram.shape[0] * feat_sgram.shape[1], feat_sgram.shape[2]))

        with PdfPages(plot_fname) as pdf:
            # page1 we always have a mixture and prediction, target optional if truth provided
            tfunc_name = mixdb.targets[0].truth_settings[0].function  # first target, assumes all have same
            if tfunc_name == 'mapped_snr_f':  # leave as unmapped snr
                predplot = predict
                tfunc_name = mixdb.targets[0].truth_settings[0].function
            elif tfunc_name == 'target_f':
                predplot = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            else:
                predplot = 10 * np.log10(predict + np.finfo(np.float32).eps)  # use db scale
                tfunc_name = tfunc_name + ' (db)'

            mixspec = 20 * np.log10(abs(mixture_f) + np.finfo(np.float32).eps)
            pdf.savefig(plot_mixpred(mixture=mixture,
                                     mixture_f=mixspec,
                                     target=target,
                                     feature=feat_sgram,
                                     predict=predplot,
                                     tp_title=tfunc_name))

            # ----- page 2, plot unmapped predict, opt truth reconstructed and line plots of mean-over-f
            # pdf.savefig(plot_pdb_predtruth(predict=pred_snr_f, tp_title='predict snr_f (db)'))

            # page 3 speech extraction
            tg_spec = 20 * np.log10(abs(target_f) + np.finfo(np.float32).eps)
            tg_est_spec = 20 * np.log10(abs(predict_complex) + np.finfo(np.float32).eps)
            # n_spec = np.reshape(n_spec,(n_spec.shape[0] * n_spec.shape[1], n_spec.shape[2]))
            pdf.savefig(plot_epredtruth(predict=tg_est_spec,
                                        predict_wav=target_est_wav,
                                        truth_f=tg_spec,
                                        truth_wav=target,
                                        metric=lerr_tg_frame,
                                        tp_title='speech estimate'))

            # page 4 noise extraction
            n_spec = 20 * np.log10(abs(noise_f) + np.finfo(np.float32).eps)
            n_est_spec = 20 * np.log10(abs(noise_est_complex) + np.finfo(np.float32).eps)
            pdf.savefig(plot_epredtruth(predict=n_est_spec,
                                        predict_wav=noise_est_wav,
                                        truth_f=n_spec,
                                        truth_wav=noisefi,
                                        metric=lerr_n_frame,
                                        tp_title='noise estimate'))

            if truth_est_mode:
                # page 5 truth-based speech extraction
                tg_trest_spec = 20 * np.log10(abs(target_truth_est_complex) + np.finfo(np.float32).eps)
                pdf.savefig(plot_epredtruth(predict=tg_trest_spec,
                                            predict_wav=target_truth_est_audio,
                                            truth_f=tg_spec,
                                            truth_wav=target,
                                            metric=lerr_tgtr_frame,
                                            tp_title='truth-based speech estimate'))

                # page 6 truth-based noise extraction
                n_trest_spec = 20 * np.log10(abs(noise_truth_est_complex) + np.finfo(np.float32).eps)
                pdf.savefig(plot_epredtruth(predict=n_trest_spec,
                                            predict_wav=noise_truth_est_audio,
                                            truth_f=n_spec,
                                            truth_wav=noisefi,
                                            metric=lerr_ntr_frame,
                                            tp_title='truth-based noise estimate'))

        plt.close('all')

    return all_metrics_table_1, all_metrics_table_2, all_metrics_table_3, all_metrics_table_4


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    predict_location = args['--ploc']
    truth_est_mode = args['--truth-est-mode']
    enable_plot = args['--plot']
    enable_wav = args['--wav']
    enable_summary = args['--summary']
    input_name = args['INPUT']

    from os.path import join

    from tqdm import tqdm

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import p_tqdm_map

    # Setup logging file
    create_file_handler(join(predict_location, 'calc_metric_spenh_targetf.log'))
    update_console_handler(verbose)
    initial_log_messages('calc_metric_spenh_targetf')

    mixdb = MixtureDatabase(config=input_name, show_progress=True)
    mixids = mixdb.mixids_to_list(mixids)
    logger.info(f'Found {len(mixids)} mixtures with {mixdb.num_classes} classes from {input_name}')

    mtab_snr_summary_tr = None
    mtab_snr_summary_emtr = None

    MP_GLOBAL.mixdb = mixdb
    MP_GLOBAL.predict_location = predict_location
    MP_GLOBAL.truth_est_mode = truth_est_mode
    MP_GLOBAL.enable_plot = enable_plot
    MP_GLOBAL.enable_wav = enable_wav

    progress = tqdm(total=len(mixids), desc='calc_metric_spenh_targetf')
    all_metrics_tables = p_tqdm_map(_process_mixture, mixids, progress=progress)
    progress.close()

    all_metrics_table_1 = pd.concat([item[0] for item in all_metrics_tables])
    all_metrics_table_2 = pd.concat([item[1] for item in all_metrics_tables])
    all_metrics_table_3 = pd.concat([item[2] for item in all_metrics_tables])
    all_metrics_table_4 = pd.concat([item[3] for item in all_metrics_tables])

    if not enable_summary:
        return

    # 9) Done with mixtures, write out summary metrics
    # Calculate SNR summary for
    all_mtab1_sorted = all_metrics_table_1.sort_values(by=['MXSNR', 'SPFILE'])
    all_mtab2_sorted = all_metrics_table_2.sort_values(by=['MXSNR'])
    mtab_snr_summary = all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(numeric_only=True).to_frame().T
    mtab_snr_summary_em = all_mtab2_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(numeric_only=True).to_frame().T
    for snri in range(1, len(mixdb.snrs)):
        mtab_snr_summary = pd.concat([mtab_snr_summary,
                                      all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                          numeric_only=True).to_frame().T])
        mtab_snr_summary_em = pd.concat([mtab_snr_summary_em,
                                         all_mtab2_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                             numeric_only=True).to_frame().T])

    if truth_est_mode:
        all_mtab3_sorted = all_metrics_table_3.sort_values(by=['MXSNR'])
        all_mtab4_sorted = all_metrics_table_4.sort_values(by=['MXSNR'])
        mtab_snr_summary_tr = all_mtab3_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(
            numeric_only=True).to_frame().T
        mtab_snr_summary_emtr = all_mtab4_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(
            numeric_only=True).to_frame().T
        for snri in range(1, len(mixdb.snrs)):
            mtab_snr_summary_tr = pd.concat([mtab_snr_summary_tr,
                                             all_mtab3_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                                 numeric_only=True).to_frame().T])
            mtab_snr_summary_emtr = pd.concat([mtab_snr_summary_emtr,
                                               all_mtab4_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                                   numeric_only=True).to_frame().T])

    num_mix = len(mixids)
    if num_mix > 1:
        with open(join(predict_location, 'metric_spenh_targetf_summary.txt'), 'w') as f:
            print(f'Speech enhancement metrics avg over each SNR:', file=f)
            print(mtab_snr_summary.round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics stats avg over each SNR:', file=f)
            print(mtab_snr_summary_em.round(2).to_string(), file=f)
            print('', file=f)

            print(f'Speech enhancement metrics stats over all {num_mix} mixtures:', file=f)
            print(all_metrics_table_1.describe().round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics stats over all {num_mix} mixtures:', file=f)
            print(all_metrics_table_2.describe().round(2).to_string(), file=f)
            print('', file=f)

            if truth_est_mode:
                print(f'Truth-based speech enhancement metrics avg over each SNR:', file=f)
                print(mtab_snr_summary_tr.round(2).to_string(), file=f)
                print('', file=f)
                print(f'Truth-based extraction statistics stats avg over each SNR:', file=f)
                print(mtab_snr_summary_emtr.round(2).to_string(), file=f)
                print('', file=f)

                print(f'Truth-based speech enhancement metrics stats over all {num_mix} mixtures:', file=f)
                print(all_metrics_table_3.describe().round(2).to_string(), file=f)
                print('', file=f)
                print(f'Truth-based extraction statistic stats over all {num_mix} mixtures:', file=f)
                print(all_metrics_table_4.describe().round(2).to_string(), file=f)
                print('', file=f)

            print('Speech enhancement metrics all-mixtures list:', file=f)
            print(all_metrics_table_1.round(2).to_string(), file=f)
            print('', file=f)
            print('Extraction statistics all-mixtures list:', file=f)
            print(all_metrics_table_2.round(2).to_string(), file=f)

            # Write summary to .csv file
            csv_name = join(predict_location, 'metric_spenh_targetf_summary.csv')
            header_args = {
                'mode': 'a',
                'encoding': 'utf-8',
                'index': False,
                'header': False,
            }
            table_args = {
                'mode': 'a',
                'encoding': 'utf-8',
            }
            label = f'Speech enhancement metrics stats over {num_mix} mixtures:'
            pd.DataFrame([label]).to_csv(csv_name, **header_args)
            all_metrics_table_1.describe().round(2).to_csv(csv_name, encoding='utf-8')
            pd.DataFrame(['']).to_csv(csv_name, **header_args)

            label = f'Extraction statistics stats over {num_mix} mixtures:'
            pd.DataFrame([label]).to_csv(csv_name, **header_args)
            all_metrics_table_2.describe().round(2).to_csv(csv_name, **table_args)
            pd.DataFrame(['']).to_csv(csv_name, **header_args)

            if truth_est_mode:
                label = 'Speech enhancement metrics of extraction method using truth, stats:'
                pd.DataFrame([label]).to_csv(csv_name, **header_args)
                all_metrics_table_3.describe().round(2).to_csv(csv_name, **table_args)
                pd.DataFrame(['']).to_csv(csv_name, **header_args)

                label = 'Truth extraction statistics stats:'
                pd.DataFrame([label]).to_csv(csv_name, **header_args)
                all_metrics_table_4.describe().round(2).to_csv(csv_name, **table_args)
                pd.DataFrame(['']).to_csv(csv_name, **header_args)

            csv_name = join(predict_location, 'metric_spenh_targetf_list.csv')
            pd.DataFrame(['Speech enhancement metrics list:']).to_csv(csv_name, **header_args)
            all_metrics_table_1.round(2).to_csv(csv_name, **table_args)

            csv_name = join(predict_location, 'metric_spenh_targetf_estats_list.csv')
            pd.DataFrame(['Extraction statistics list:']).to_csv(csv_name, **header_args)
            all_metrics_table_2.round(2).to_csv(csv_name, **table_args)

            if truth_est_mode:
                csv_name = join(predict_location, 'metric_spenh_targetf_truth_list.csv')
                pd.DataFrame(['Speech enhancement metrics list:']).to_csv(csv_name, **header_args)
                all_metrics_table_3.round(2).to_csv(csv_name, **table_args)

                csv_name = join(predict_location, 'metric_spenh_targetf_truth_list.csv')
                pd.DataFrame(['Extraction statistics list:']).to_csv(csv_name, **header_args)
                all_metrics_table_4.round(2).to_csv(csv_name, **table_args)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



"""sonusai gentcst

usage: gentcst [-hvru] [-o OUTPUT] [-f FOLD] [-y ONTOLOGY]

options:
    -h, --help
    -v, --verbose                       Be verbose.
    -o OUTPUT, --output OUTPUT          Output file name.
    -f FOLD, --fold FOLD                Fold(s) to include. [default: *].
    -y ONTOLOGY, --ontology ONTOLOGY    Reference ontology JSON file for cross-check and adding metadata. [default: ontology.json].
    -r, --hierarchical                  Generate hierarchical multiclass truth for non-leaf nodes.
    -u, --update                        Write JSON metadata into the tree.

Generate a target SonusAI configuration file from a hierarchical subdirectory tree
under the local directory. Leaves in the subdirectory tree define classes by providing SonusAI list
.txt files for class data and additional truth config definitions. This offers a way to simplify and
more clearly manage hierarchical class structure similar to the Audioset ontology.

Features:
    - automatically compiles the target configuration with the number of classes and each truth
      index and the associated files (note: files should use absolute path with env variable)
    - supports hierarchical multilabel class truth generation, where higher nodes in the tree
      are included with leaf truth index, and different leaves with the same name will have the
      same index (i.e. door/squeak, chair/squeak, and brief-tone/squeak)
    - support config overrides for class config.yml or filename.yml for specific file config
    - manages folds via file prefix naming convention, i.e. 1-*.txt, 2-*.txt, etc.
      specifies data for fold 1, 2, etc.
    - cross-check node definitions with the reference ontology and auto copy class metadata into
      the tree to help define/collect data
    - class label is also generated (in the .yml or referencing a .csv)

Inputs:
    The local subdirectory tree with expected top-level configuration file config.yml. This
    file must at least define the feature and can have additional parameters treated as default
    (for truth, augmentation, etc.) which will be included in the generated output yaml file.
    An entire class dataset config parameter can be overridden by including it in a config.yml
    file in the class subdirectory. Individual file config parameters can be further overridden
    or specified in a file of the same name but with .yml (i.e. 1-data.yml with the 1-data.txt)

Outputs:
    output.yml
    output.csv
    gentcst.log

"""
from typing import Dict
from typing import List

from sonusai import logger

CONFIG_FILE = 'config.yml'


def gentcst(fold: str = '*',
            ontology: str = 'ontology.json',
            hierarchical: bool = False,
            update: bool = False,
            verbose: bool = False) -> (Dict, List[Dict]):
    import os
    from copy import deepcopy
    from os.path import dirname
    from os.path import exists
    from os.path import splitext

    from sonusai import SonusAIError
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import get_default_config
    from sonusai.mixture import raw_load_config
    from sonusai.mixture import update_config_from_hierarchy
    from sonusai.mixture import update_truth_settings

    update_console_handler(verbose)
    initial_log_messages('gentcst')

    if update:
        logger.info('Updating tree with JSON metadata')
        logger.info('')

    logger.debug(f'fold:         {fold}')
    logger.debug(f'ontology:     {ontology}')
    logger.debug(f'hierarchical: {hierarchical}')
    logger.debug(f'update:       {update}')
    logger.debug('')

    config = get_config()

    if config['truth_mode'] == 'mutex' and hierarchical:
        raise SonusAIError('Multi-class truth is incompatible with truth_mode mutex')

    all_files = get_all_files(hierarchical)
    all_folds = get_folds_from_files(all_files)
    use_folds = get_use_folds(all_folds, fold)
    use_files = get_files_from_folds(all_files, use_folds)
    report_leaf_fold_data_usage(all_files, use_files)

    ontology_data = validate_ontology(ontology, use_files)

    labels = get_labels_from_files(all_files)
    logger.debug('Truth indices:')
    for item in labels:
        logger.debug(f' {item["index"]:3} {item["display_name"]}')
    logger.debug('')

    gen_truth_indices(use_files, labels)

    config['num_classes'] = len(labels)
    if config['truth_mode'] == 'mutex':
        config['num_classes'] = config['num_classes'] + 1

    config['targets'] = []

    logger.info(f'gentcst {len(use_files)} entries in tree')
    root = os.getcwd()
    for file in use_files:
        leaf = dirname(file['file'])
        local_config = get_default_config()
        local_config = update_config_from_hierarchy(root=root, leaf=leaf, config=local_config)
        if not isinstance(local_config['truth_settings'], list):
            local_config['truth_settings'] = [local_config['truth_settings']]

        specific_name = splitext(file['file'])[0] + '.yml'
        if exists(specific_name):
            specific_config = raw_load_config(specific_name)

            for key in specific_config:
                if key != 'truth_settings':
                    local_config[key] = specific_config[key]
                else:
                    local_config['truth_settings'] = update_truth_settings(given=specific_config['truth_settings'],
                                                                           default=local_config['truth_settings'])

        truth_settings = deepcopy(local_config['truth_settings'])
        for idx, val in enumerate(truth_settings):
            val['index'] = file['truth_index']
            for key in ['function', 'config']:
                if key in val and val[key] == config['truth_settings'][idx][key]:
                    del val[key]

        target = {
            'name': file['file'],
            'truth_settings': truth_settings
        }

        config['targets'].append(target)

        if update:
            write_metadata_to_tree(ontology_data, file)

    return config, labels


def get_ontology(ontology: str) -> list:
    import json
    from os.path import exists

    ontology_data = []

    if exists(ontology):
        with open(file=ontology, mode='r', encoding='utf-8') as f:
            ontology_data = json.load(f)

    # Convert names to lowercase, convert spaces to '-', split into lists on ','
    for item in ontology_data:
        name = item['name'].lower()
        # Need to find ', ' before converting spaces so that we don't convert ', ' to ',-'
        name = name.replace(', ', ',')
        name = name.replace(' ', '-')
        name = name.split(',')
        item['name'] = name

    return ontology_data


def get_dirs() -> list:
    import sh

    match = sh.find('.', '-type', 'd')
    match = list(map(str.strip, match))

    dirs = []
    for m in match:
        if not m.startswith('./gentcst') and m != '.':
            classes = m.split('/')
            # Remove first element because it is '.'
            classes.pop(0)

            dirs.append({'file': m, 'classes': classes})

    return dirs


def get_leaf_from_classes(classes: list) -> str:
    return '/'.join(classes)


def get_all_files(hierarchical: bool) -> list:
    import re

    import sh

    match = sh.find('.', '-regex', '.*/[0-9]+-.*\.txt')
    match = list(map(str.strip, match))

    files = []
    fold_pattern = re.compile('.*/(\d+)-.*\.txt')
    for m in match:
        fold_match = fold_pattern.match(m)
        if fold_match:
            fold = int(fold_match.group(1))

            classes = m.split('/')
            # Remove first element because it is '.'
            classes.pop(0)
            # Remove last element because it is the file name
            classes.pop()

            leaf = get_leaf_from_classes(classes)

            if hierarchical:
                labels = classes
            else:
                labels = [leaf]

            files.append({'file': m, 'classes': classes, 'leaf': leaf, 'labels': labels, 'fold': fold})

    return files


def get_use_folds(all_folds: list, fold: str) -> list:
    import numpy as np

    req_folds = get_req_folds(all_folds, fold)
    use_folds = list(set(req_folds).intersection(all_folds))

    logger.debug('Fold information')
    logger.debug(f' Available: {all_folds}')
    logger.debug(f' Requested: {req_folds}')
    logger.debug(f' Used:      {use_folds}')
    logger.debug(f' Unused:    {list(set(use_folds).symmetric_difference(all_folds))}')
    logger.debug(f' Missing:   {list(np.setdiff1d(req_folds, all_folds))}')
    logger.debug('')

    return use_folds


def get_files_from_folds(files: list, folds: list) -> list:
    use_files = []
    for item in files:
        if item['fold'] in folds:
            use_files.append(item)
    return use_files


def get_item_from_name(ontology_data: list, name: str) -> dict:
    return next((item for item in ontology_data if name in item['name']), None)


def get_item_from_id(ontology_data: list, identity: str) -> dict:
    return next((item for item in ontology_data if item['id'] == identity), None)


def get_name_from_id(ontology_data: list, identity: str) -> list:
    from sonusai import SonusAIError

    name = None
    found = False

    for item in ontology_data:
        if item['id'] == identity:
            if found:
                raise SonusAIError(f'id {identity} appears multiple times in ontology')

            name = item['name']
            found = True

    return name


def get_id_from_name(ontology_data: list, name: str) -> str:
    from sonusai import SonusAIError

    identity = None
    found = False

    for item in ontology_data:
        if name in item['name']:
            if found:
                raise SonusAIError(f'name {name} appears multiple times in ontology')

            identity = item['id']
            found = True

    return identity


def is_valid_name(ontology_data: list, name: str) -> bool:
    for item in ontology_data:
        if name in item['name']:
            return True
    return False


def is_valid_child(ontology_data: list, parent: str, child: str) -> bool:
    valid = False
    parent_item = get_item_from_name(ontology_data, parent)
    child_id = get_id_from_name(ontology_data, child)

    if child_id is not None and parent_item is not None:
        if child_id in parent_item['child_ids']:
            valid = True

    return valid


def is_valid_hierarchy(ontology_data: list, classes: list) -> bool:
    valid = True

    for parent, child in zip(classes, classes[1:]):
        if not is_valid_child(ontology_data, parent, child):
            valid = False

    return valid


def validate_class(ontology_data: list, file: dict) -> bool:
    valid = True
    classes = file['classes']

    for c in classes:
        if not is_valid_name(ontology_data, c):
            logger.warning(f' Could not find {c} in ontology for {file["file"]}')
            valid = False

    if valid:
        if not is_valid_hierarchy(ontology_data, classes):
            logger.warning(f' Invalid parent/child relationship for {file["file"]}')
            valid = False

    return valid


def get_req_folds(folds: list, fold: str) -> list:
    from sonusai.utils import expand_range

    if fold == '*':
        return folds

    return expand_range(fold)


def get_folds_from_files(files: list) -> list:
    result = [file['fold'] for file in files]
    # Converting to set and back to list ensures uniqueness
    return sorted(list(set(result)))


def get_leaves_from_files(files: list) -> list:
    result = [file['leaf'] for file in files]
    # Converting to set and back to list ensures uniqueness
    return sorted(list(set(result)))


def get_labels_from_files(files: List[Dict]) -> List[Dict]:
    all_labels = [file['labels'] for file in files]

    # Get labels by depth
    labels_by_depth = [[] for _ in range(max([len(label) for label in all_labels]))]
    for label in all_labels:
        for index, name in enumerate(label):
            labels_by_depth[index].append(name)

    for n in range(len(labels_by_depth)):
        # Converting to set and back to list ensures uniqueness
        labels_by_depth[n] = sorted(list(set(labels_by_depth[n])))

    # We want the deepest leaves first
    labels_by_depth.reverse()

    # Now flatten the list
    labels_by_depth = [item for sublist in labels_by_depth for item in sublist]

    # Generate index, name pairs
    labels = []
    for index, file in enumerate(labels_by_depth):
        labels.append({'index': index + 1, 'display_name': file})

    return labels


def gen_truth_indices(files: list, labels: list):
    for file in files:
        file['truth_index'] = []
        for label in file['labels']:
            file['truth_index'].append(get_index_from_label(labels, label))
        file['truth_index'] = sorted(file['truth_index'])


def get_index_from_label(labels: list, label: str) -> int:
    return next((item for item in labels if item['display_name'] == label), None)['index']


def get_config() -> Dict:
    from os.path import exists

    from sonusai import SonusAIError
    from sonusai.mixture import raw_load_config

    if not exists(CONFIG_FILE):
        raise SonusAIError(f'No {CONFIG_FILE} at top level')

    config = raw_load_config(CONFIG_FILE)

    if 'feature' not in config:
        raise SonusAIError('feature not in top level config')

    if 'truth_mode' not in config:
        raise SonusAIError('truth_mode not in top level config')

    if config['truth_mode'] not in ['normal', 'mutex']:
        raise SonusAIError('Invalid truth_mode in top level config')

    if 'truth_settings' in config and not isinstance(config['truth_settings'], list):
        config['truth_settings'] = [config['truth_settings']]

    return config


def validate_ontology(ontology: str, files: list) -> list:
    from os.path import exists

    if exists(ontology):
        ontology_data = get_ontology(ontology)
        logger.debug(f'Reference ontology in {ontology} has {len(ontology_data)} classes')
        logger.debug('')

        logger.info('Checking tree against reference ontology')
        all_dirs = get_dirs()
        valid = True
        for file in all_dirs:
            if not validate_class(ontology_data, file):
                valid = False
        if valid:
            logger.info('PASS')
        logger.info('')

        logger.info('Checking files against reference ontology')
        valid = True
        for file in files:
            if not validate_class(ontology_data, file):
                valid = False
        if valid:
            logger.info('PASS')
        logger.info('')

        return ontology_data

    return []


def get_node_from_name(ontology_data: list, name: str) -> dict:
    nodes = [item for item in ontology_data if name in item['name']]
    if len(nodes) == 1:
        return nodes[0]

    if nodes:
        logger.warning(f'Found multiple entries in reference ontology that match {name}')
    else:
        logger.warning(f'Could not find entry for {name} in reference ontology')

    return {}


def write_metadata_to_tree(ontology_data: list, file: dict):
    import json
    from os.path import basename
    from os.path import dirname

    if ontology_data:
        node = get_node_from_name(ontology_data, file['classes'][-1])
        if node:
            dir_name = dirname(file['file'])
            json_name = dir_name + '/' + basename(dir_name) + '.json'
            with open(file=json_name, mode='w') as f:
                json.dump(node, f)


def get_folds_from_leaf(all_files: list, leaf: str) -> list:
    files = [item for item in all_files if item['leaf'] == leaf]
    return get_folds_from_files(files)


def report_leaf_fold_data_usage(all_files: list, use_files: list):
    use_leaves = get_leaves_from_files(use_files)
    all_leaves = get_leaves_from_files(all_files)

    logger.debug('Data folds present in each leaf')
    leaf_len = len(max(all_leaves, key=len))
    for leaf in all_leaves:
        folds = get_folds_from_leaf(all_files, leaf)
        logger.debug(f' {leaf:{leaf_len}} {folds}')
    logger.debug('')

    dif_leaves = set(all_leaves).symmetric_difference(use_leaves)
    if dif_leaves:
        logger.warning('This fold selection is missing data from the following leaves')
        for c in dif_leaves:
            logger.warning(f' {c}')
        logger.warning('')


def main():
    import csv
    from os import getcwd
    from os.path import basename
    from os.path import splitext

    import yaml
    from docopt import docopt

    import sonusai
    from sonusai import create_file_handler
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    output_name = args['--output']
    fold = args['--fold']
    ontology = args['--ontology']
    hierarchical = args['--hierarchical']
    update = args['--update']

    if not output_name:
        output_name = basename(getcwd()) + '.yml'

    create_file_handler('gentcst.log')

    config, labels = gentcst(fold=fold,
                             ontology=ontology,
                             hierarchical=hierarchical,
                             update=update,
                             verbose=verbose)

    with open(file=output_name, mode='w') as f:
        yaml.dump(config, f)
        logger.info(f'Wrote config to {output_name}')

    csv_fields = ['index', 'display_name']
    csv_name = splitext(output_name)[0] + '.csv'

    with open(file=csv_name, mode='w') as f:
        writer = csv.DictWriter(f, fieldnames=csv_fields)
        writer.writeheader()
        writer.writerows(labels)
        logger.info(f'Wrote labels to {csv_name}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from dataclasses import dataclass
from os import PathLike
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import numpy.typing as npt
from dataclasses_json import DataClassJsonMixin

AudioT = npt.NDArray[np.float32]
AudiosT = List[AudioT]

ListAudiosT = List[List[AudioT]]

Truth = npt.NDArray[np.float32]
Segsnr = npt.NDArray[np.float32]

AudioF = npt.NDArray[np.complex64]
AudiosF = List[AudioF]

Feature = npt.NDArray[np.float32]

Predict = npt.NDArray[np.float32]

Location = Union[str, bytes, PathLike]


class DataClassSonusAIMixin(DataClassJsonMixin):
    from typing import Dict
    from typing import Union

    # Json type defined to maintain compatibility with DataClassJsonMixin
    Json = Union[dict, list, str, int, float, bool, None]

    def __str__(self):
        return f'{self.to_dict()}'

    # Override DataClassJsonMixin to remove dictionary keys with values of None
    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        def del_none(d):
            if isinstance(d, dict):
                for key, value in list(d.items()):
                    if value is None:
                        del d[key]
                    elif isinstance(value, dict):
                        del_none(value)
                    elif isinstance(value, list):
                        for item in value:
                            del_none(item)
            elif isinstance(d, list):
                for item in d:
                    del_none(item)
            return d

        return del_none(super().to_dict(encode_json))


@dataclass(frozen=True)
class TruthSetting(DataClassSonusAIMixin):
    config: Optional[dict] = None
    function: Optional[str] = None
    index: Optional[List[int]] = None


TruthSettings = List[TruthSetting]
OptionalNumberStr = Optional[Union[float, int, str]]
OptionalListNumberStr = Optional[List[Union[float, int, str]]]


@dataclass
class Augmentation(DataClassSonusAIMixin):
    normalize: OptionalNumberStr = None
    pitch: OptionalNumberStr = None
    tempo: OptionalNumberStr = None
    gain: OptionalNumberStr = None
    eq1: OptionalListNumberStr = None
    eq2: OptionalListNumberStr = None
    eq3: OptionalListNumberStr = None
    lpf: OptionalNumberStr = None
    ir: OptionalNumberStr = None
    count: Optional[int] = None
    mixup: Optional[int] = 1


Augmentations = List[Augmentation]


@dataclass(frozen=True)
class TargetFile(DataClassSonusAIMixin):
    duration: float
    name: Location
    truth_settings: TruthSettings
    augmentations: Optional[Augmentations] = None
    class_balancing_augmentation: Optional[Augmentation] = None


TargetFiles = List[TargetFile]


@dataclass
class AugmentedTarget(DataClassSonusAIMixin):
    target_augmentation_index: int
    target_file_index: int


AugmentedTargets = List[AugmentedTarget]


@dataclass(frozen=True)
class NoiseFile(DataClassSonusAIMixin):
    name: Location
    duration: float
    augmentations: Optional[Augmentations] = None


NoiseFiles = List[NoiseFile]
ClassCount = List[int]

GeneralizedIDs = Union[str, int, List[int], range]


@dataclass(frozen=True)
class TruthFunctionConfig(DataClassSonusAIMixin):
    feature: str
    mutex: bool
    num_classes: int
    target_gain: float
    config: Optional[dict] = None
    function: Optional[str] = None
    index: Optional[List[int]] = None


@dataclass
class GenMixData:
    targets: AudiosT = None
    noise: AudioT = None
    mixture: AudioT = None
    truth_t: Optional[Truth] = None
    segsnr_t: Optional[Segsnr] = None


@dataclass
class GenFTData:
    feature: Optional[Feature] = None
    truth_f: Optional[Truth] = None
    segsnr: Optional[Segsnr] = None


@dataclass(frozen=True)
class ImpulseResponseData:
    sample_rate: int
    offset: int
    filter: List[np.float32]


ImpulseResponseFiles = List[Location]



def sample_to_dbfs(sample: int, bits: int = 16) -> float:
    import numpy as np

    return 20 * np.log10(abs(sample) / (2 ** (bits - 1)))


def dbfs_to_sample(dbfs: float, bits: int = 16) -> int:
    return int((10 ** (dbfs / 20)) * (2 ** (bits - 1)))



import numpy as np

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import Feature


def get_feature_from_audio(audio: AudioT, feature: str) -> Feature:
    from sonusai.mixture import MRecord
    from sonusai.mixture import MixtureDatabaseConfig
    from sonusai.mixture import get_pad_length

    mixdb = MixtureDatabase(config=MixtureDatabaseConfig(feature=feature,
                                                         num_classes=1,
                                                         truth_mutex=False,
                                                         truth_reduction_function='max'),
                            lazy_load=True)

    audio = np.pad(array=audio, pad_width=(0, get_pad_length(len(audio), mixdb.feature_step_samples)))
    mixdb.mixtures = [MRecord(samples=len(audio))]

    data, _ = mixdb.mixture_ft(mixid=0, mixture=audio, truth_t=np.empty((len(audio), 1)), force=True)
    return data



"""sonusai genft

usage: genft [-hvs] [-i MIXID] LOC

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -s, --segsnr                    Save segsnr. [default: False].

Generate SonusAI feature/truth data from a SonusAI mixture database.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.

Outputs the following to the mixture database directory:
    <id>.h5:
        dataset:    feature
        dataset:    truth_f
        dataset:    segsnr (optional)
    <id>.txt
    genft.log

"""
from dataclasses import dataclass
from typing import List

from sonusai import logger
from sonusai.mixture import GenFTData
from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import MixtureDatabase


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    compute_truth: bool = None
    compute_segsnr: bool = None
    force: bool = None
    write: bool = None


MP_GLOBAL = MPGlobal()


def genft(mixdb: MixtureDatabase,
          mixids: GeneralizedIDs = None,
          compute_truth: bool = True,
          compute_segsnr: bool = False,
          write: bool = False,
          show_progress: bool = False,
          force: bool = True) -> List[GenFTData]:
    import multiprocessing as mp

    from tqdm import tqdm

    from sonusai.utils import p_tqdm_map

    MP_GLOBAL.mixdb = mixdb
    MP_GLOBAL.compute_truth = compute_truth
    MP_GLOBAL.compute_segsnr = compute_segsnr
    MP_GLOBAL.force = force
    MP_GLOBAL.write = write

    mixids = mixdb.mixids_to_list(mixids)
    if mp.current_process().daemon:
        results = []
        for mixid in mixids:
            results.append(_genft_kernel(mixid))
    else:
        progress = tqdm(total=len(mixids), desc='genft', disable=not show_progress)
        results = p_tqdm_map(_genft_kernel, mixids, progress=progress)
        progress.close()

    return results


def _genft_kernel(mixid: int) -> GenFTData:
    feature, truth_f = MP_GLOBAL.mixdb.mixture_ft(mixid=mixid, force=MP_GLOBAL.force)
    write_data = [('feature', feature)]

    if MP_GLOBAL.compute_truth:
        write_data.append(('truth_f', truth_f))
    else:
        truth_f = None

    if MP_GLOBAL.compute_segsnr:
        segsnr = MP_GLOBAL.mixdb.mixture_segsnr(mixid=mixid, force=MP_GLOBAL.force)
        write_data.append(('segsnr', segsnr))
    else:
        segsnr = None

    if MP_GLOBAL.write:
        MP_GLOBAL.mixdb.write_mixture_data(mixid, write_data)
        MP_GLOBAL.mixdb.write_mixture_metadata(mixid)

    return GenFTData(feature=feature, truth_f=truth_f, segsnr=segsnr)


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import time
    from os.path import join

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms

    verbose = args['--verbose']
    mixids = args['--mixid']
    compute_segsnr = args['--segsnr']
    location = args['LOC']

    start_time = time.monotonic()

    create_file_handler(join(location, 'genft.log'))
    update_console_handler(verbose)
    initial_log_messages('genft')

    logger.info(f'Load mixture database from {location}')
    mixdb = MixtureDatabase(config=location, show_progress=True)
    mixids = mixdb.mixids_to_list(mixids)

    total_samples = mixdb.total_samples(mixids)
    duration = total_samples / sonusai.mixture.SAMPLE_RATE
    total_transform_frames = total_samples // mixdb.ft_config.R
    total_feature_frames = total_samples // mixdb.feature_step_samples

    logger.info('')
    logger.info(f'Found {len(mixids):,} mixtures to process')
    logger.info(f'{total_samples:,} samples, '
                f'{total_transform_frames:,} transform frames, '
                f'{total_feature_frames:,} feature frames')

    mixdb.check_audio_files_exist()

    genft(mixdb=mixdb,
          mixids=mixids,
          compute_segsnr=compute_segsnr,
          write=True,
          show_progress=True)

    logger.info(f'Wrote {len(mixids)} mixtures to {location}')
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'feature:  {human_readable_size(total_feature_frames * mixdb.fg.stride * mixdb.fg.num_bands * 4, 1)}')
    logger.info(f'truth_f:  {human_readable_size(total_feature_frames * mixdb.num_classes * 4, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(total_transform_frames * 4, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')
    logger.info('')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



"""sonusai genmix

usage: genmix [-hvgts] [-i MIXID] LOC

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -g, --target                    Save target. [default: False].
    -t, --truth                     Save truth_t. [default: False].
    -s, --segsnr                    Save segsnr_t. [default: False].

Generate SonusAI mixture data from a SonusAI mixture database.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.

Outputs the following to the mixture database directory:
    <id>.h5:
        dataset:    mixture
        dataset:    targets
        dataset:    noise
        dataset:    target (optional)
        dataset:    truth_t (optional)
        dataset:    segsnr_t (optional)
    <id>.txt
    genmix.log
"""
from dataclasses import dataclass
from typing import List

from sonusai import logger
from sonusai.mixture import GenMixData
from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import MixtureDatabase


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    save_target: bool = None
    compute_truth: bool = None
    compute_segsnr: bool = None
    force: bool = None
    write: bool = None


MP_GLOBAL = MPGlobal()


def genmix(mixdb: MixtureDatabase,
           mixids: GeneralizedIDs = None,
           save_target: bool = False,
           compute_truth: bool = False,
           compute_segsnr: bool = False,
           write: bool = False,
           show_progress: bool = False,
           force: bool = True) -> List[GenMixData]:
    import multiprocessing as mp

    from tqdm import tqdm

    from sonusai.utils import p_tqdm_map

    MP_GLOBAL.mixdb = mixdb
    MP_GLOBAL.save_target = save_target
    MP_GLOBAL.compute_truth = compute_truth
    MP_GLOBAL.compute_segsnr = compute_segsnr
    MP_GLOBAL.force = force
    MP_GLOBAL.write = write

    mixids = mixdb.mixids_to_list(mixids)
    if mp.current_process().daemon:
        results = []
        for mixid in mixids:
            results.append(_genmix_kernel(mixid))
    else:
        progress = tqdm(total=len(mixids), desc='genmix', disable=not show_progress)
        results = p_tqdm_map(_genmix_kernel, mixids, progress=progress)
        progress.close()

    return results


def _genmix_kernel(mixid: int) -> GenMixData:
    targets = MP_GLOBAL.mixdb.mixture_targets(mixid=mixid, force=MP_GLOBAL.force)
    noise = MP_GLOBAL.mixdb.mixture_noise(mixid=mixid, force=MP_GLOBAL.force)
    write_data = [('targets', targets), ('noise', noise)]

    if MP_GLOBAL.compute_truth:
        truth_t = MP_GLOBAL.mixdb.mixture_truth_t(mixid=mixid, targets=targets, noise=noise, force=MP_GLOBAL.force)
        write_data.append(('truth_t', truth_t))
    else:
        truth_t = None

    target = MP_GLOBAL.mixdb.mixture_target(mixid=mixid, targets=targets)
    if MP_GLOBAL.save_target:
        write_data.append(('target', target))

    if MP_GLOBAL.compute_segsnr:
        segsnr_t = MP_GLOBAL.mixdb.mixture_segsnr_t(mixid=mixid,
                                                    targets=targets,
                                                    target=target,
                                                    noise=noise,
                                                    force=MP_GLOBAL.force)
        write_data.append(('segsnr_t', segsnr_t))
    else:
        segsnr_t = None

    mixture = MP_GLOBAL.mixdb.mixture_mixture(mixid=mixid,
                                              targets=targets,
                                              target=target,
                                              noise=noise,
                                              force=MP_GLOBAL.force)
    write_data.append(('mixture', mixture))

    if MP_GLOBAL.write:
        MP_GLOBAL.mixdb.write_mixture_data(mixid, write_data)
        MP_GLOBAL.mixdb.write_mixture_metadata(mixid)

    return GenMixData(targets=targets,
                      noise=noise,
                      mixture=mixture,
                      truth_t=truth_t,
                      segsnr_t=segsnr_t)


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import time
    from os.path import join

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms

    verbose = args['--verbose']
    location = args['LOC']
    mixids = args['--mixid']
    save_target = args['--target']
    compute_truth = args['--truth']
    compute_segsnr = args['--segsnr']

    start_time = time.monotonic()

    create_file_handler(join(location, 'genmix.log'))
    update_console_handler(verbose)
    initial_log_messages('genmix')

    logger.info(f'Load mixture database from {location}')
    mixdb = MixtureDatabase(config=location, show_progress=True)
    mixids = mixdb.mixids_to_list(mixids)

    total_samples = mixdb.total_samples(mixids)
    duration = total_samples / sonusai.mixture.SAMPLE_RATE

    logger.info('')
    logger.info(f'Found {len(mixids):,} mixtures to process')
    logger.info(f'{total_samples:,} samples')

    mixdb.check_audio_files_exist()

    genmix(mixdb=mixdb,
           mixids=mixids,
           save_target=save_target,
           compute_truth=compute_truth,
           compute_segsnr=compute_segsnr,
           write=True,
           show_progress=True)

    logger.info(f'Wrote {len(mixids)} mixtures to {location}')
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'mixture:  {human_readable_size(total_samples * 2, 1)}')
    if compute_truth:
        logger.info(f'truth_t:  {human_readable_size(total_samples * mixdb.num_classes * 4, 1)}')
    logger.info(f'target:   {human_readable_size(total_samples * 2, 1)}')
    logger.info(f'noise:    {human_readable_size(total_samples * 2, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(total_samples * 4, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')
    logger.info('')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import AugmentedTargets


def balance_targets(mixdb: MixtureDatabase, augmented_targets: AugmentedTargets) -> AugmentedTargets:
    import numpy as np

    from sonusai.mixture import get_mixups
    from sonusai.mixture import AugmentedTarget
    from sonusai.mixture import get_augmented_target_indices_by_class

    if not mixdb.class_balancing:
        return augmented_targets

    mixups = get_mixups(mixdb.target_augmentations)
    if 1 in mixups:
        mixups.remove(1)

    for mixup in mixups:
        augmented_target_indices_by_class = get_augmented_target_indices_by_class(
            mixdb=mixdb,
            augmented_targets=augmented_targets,
            mixup=mixup)

        largest = max([len(item) for item in augmented_target_indices_by_class])
        largest = int(np.ceil(np.float32(largest) / mixup)) * mixup
        for at_indices in augmented_target_indices_by_class:
            additional_augmentations_needed = largest - len(at_indices)
            target_file_indices = sorted(
                list(set([augmented_targets[at_index].target_file_index for at_index in at_indices])))

            tfi_idx = 0
            for _ in range(additional_augmentations_needed):
                target_file_index = target_file_indices[tfi_idx]
                tfi_idx = (tfi_idx + 1) % len(target_file_indices)
                augmentation_index = _get_unused_balancing_augmentation(mixdb=mixdb,
                                                                        augmented_targets=augmented_targets,
                                                                        target_file_index=target_file_index,
                                                                        mixup=mixup)
                augmented_target = AugmentedTarget(target_file_index=target_file_index,
                                                   target_augmentation_index=augmentation_index)
                augmented_targets.append(augmented_target)

    return augmented_targets


def _get_unused_balancing_augmentation(mixdb: MixtureDatabase,
                                       augmented_targets: AugmentedTargets,
                                       target_file_index: int,
                                       mixup: int) -> int:
    """Get an unused balancing augmentation for a given target file index."""
    from dataclasses import asdict

    from sonusai.mixture import get_augmentations
    from sonusai.mixture import get_class_balancing_augmentation

    balancing_augmentations = [item for item in range(len(mixdb.target_augmentations)) if
                               item >= mixdb.first_cba_index]
    used_balancing_augmentations = [at.target_augmentation_index for at in augmented_targets if
                                    at.target_file_index == target_file_index and
                                    at.target_augmentation_index in balancing_augmentations]

    augmentation_indices = [item for item in balancing_augmentations if
                            item not in used_balancing_augmentations and
                            mixdb.target_augmentations[item].mixup == mixup]
    class_balancing_augmentation = get_class_balancing_augmentation(mixdb=mixdb,
                                                                    target_file_index=target_file_index)
    if len(augmentation_indices) == 0:
        new_augmentation = get_augmentations(rules=asdict(class_balancing_augmentation),
                                             num_ir=len(mixdb.ir_files))[0]
        new_augmentation.mixup = mixup
        mixdb.target_augmentations.append(new_augmentation)
        augmentation_indices.append(len(mixdb.target_augmentations) - 1)

    return augmentation_indices[0]



"""sonusai keras_predict

usage: keras_predict [-hvr] [-i MIXID] (-m MODEL) (-w KMODEL) [-b BATCH] [-t TSTEPS] INPUT

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate if input is a mixture database. [default: *].
    -m MODEL, --model MODEL         Python model file.
    -w KMODEL, --weights KMODEL     Keras model weights file.
    -b BATCH, --batch BATCH         Batch size.
    -t TSTEPS, --tsteps TSTEPS      Timesteps.
    -r, --reset                     Reset model between each file.

Run prediction on a trained Keras model defined by a SonusAI Keras Python model file using SonusAI genft or WAV data.

Inputs:
    MODEL       A SonusAI Python model file with build and/or hypermodel functions.
    KMODEL      A Keras model weights file (or model file with weights).
    INPUT       The input data must be one of the following:
                * WAV
                  Using the given model, generate feature data and run prediction. A model file must be
                  provided. The MIXID is ignored.

                * directory
                  Using the given SonusAI mixture database directory, generate feature and truth data if not found.
                  Run prediction. The MIXID is required.

Outputs the following to kpredict-<TIMESTAMP> directory:
    <id>.h5
        dataset:    predict
    keras_predict.log

"""
from typing import Any

from sonusai import logger
from sonusai.mixture import Feature
from sonusai.mixture import Predict


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    model_name = args['--model']
    weights_name = args['--weights']
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    reset = args['--reset']
    input_name = args['INPUT']

    from os import makedirs
    from os.path import basename
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    import h5py
    import keras_tuner as kt
    import tensorflow as tf
    from keras import backend as kb

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.data_generator import KerasFromH5
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import get_feature_from_audio
    from sonusai.mixture import read_audio
    from sonusai.utils import check_keras_overrides
    from sonusai.utils import create_ts_name
    from sonusai.utils import get_frames_per_batch
    from sonusai.utils import import_keras_model
    from sonusai.utils import reshape_outputs

    model_base = basename(model_name)

    if batch_size is not None:
        batch_size = int(batch_size)

    if timesteps is not None:
        timesteps = int(timesteps)

    output_dir = create_ts_name('kpredict')
    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    create_file_handler(join(output_dir, 'keras_predict.log'))
    update_console_handler(verbose)
    initial_log_messages('keras_predict')

    logger.info(f'tensorflow    {tf.__version__}')
    logger.info(f'keras         {tf.keras.__version__}')
    logger.info('')

    # Import model definition file
    logger.info(f'Importing {model_base}')
    model = import_keras_model(model_name)
    feature_mode = None
    num_classes = None
    with h5py.File(weights_name, 'r') as f:
        if 'sonusai_feature' in f.attrs:
            feature_mode = f.attrs['sonusai_feature']
        else:
            logger.warn(f'Could not find SonusAI feature in weights file; using hypermodel default.')
        if 'sonusai_num_classes' in f.attrs:
            num_classes = int(f.attrs['sonusai_num_classes'])
        else:
            logger.warn(f'Could not find SonusAI num_classes in weights file; using hypermodel default.')

    # Check overrides
    timesteps = check_keras_overrides(model, feature_mode, num_classes, timesteps, batch_size)

    logger.info('Building model')
    try:
        hypermodel = model.MyHyperModel(feature=feature_mode,
                                        num_classes=num_classes,
                                        timesteps=timesteps,
                                        batch_size=batch_size)
        built_model = hypermodel.build_model(kt.HyperParameters())
    except Exception as e:
        logger.exception(f'build_model() in {model_base} failed: {e}.')
        raise SystemExit(1)

    frames_per_batch = get_frames_per_batch(hypermodel.batch_size, hypermodel.timesteps)

    kb.clear_session()
    logger.info('')
    built_model.summary(print_fn=logger.info)
    logger.info('')
    logger.info(f'feature       {hypermodel.feature}')
    logger.info(f'num_classes   {hypermodel.num_classes}')
    logger.info(f'batch_size    {hypermodel.batch_size}')
    logger.info(f'timesteps     {hypermodel.timesteps}')
    logger.info(f'flatten       {hypermodel.flatten}')
    logger.info(f'add1ch        {hypermodel.add1ch}')
    logger.info(f'truth_mutex   {hypermodel.truth_mutex}')
    logger.info(f'input_shape   {hypermodel.input_shape}')
    logger.info('')

    logger.info(f'Loading weights from {weights_name}')
    built_model.load_weights(weights_name)

    if splitext(input_name)[1] == '.wav':
        # Convert WAV to feature data
        logger.info('')
        logger.info(f'Run prediction on {input_name}')
        audio = read_audio(input_name)
        feature = get_feature_from_audio(audio=audio, feature=feature_mode)

        feature, predict = pad_and_predict(hypermodel=hypermodel,
                                           built_model=built_model,
                                           feature=feature,
                                           frames_per_batch=frames_per_batch)

        output_name = splitext(input_name)[0] + '.h5'
        with h5py.File(output_name, 'a') as f:
            if 'feature' in f:
                del f['feature']
            f.create_dataset(name='feature', data=feature)

            if 'predict' in f:
                del f['predict']
            f.create_dataset(name='predict', data=predict)

        logger.info(f'Saved results to {output_name}')
        return

    if not isdir(input_name):
        logger.exception(f'Do not know how to process input from {input_name}')
        raise SystemExit(1)

    logger.info('')
    logger.info(f'Load mixture database from {input_name}')
    mixdb = MixtureDatabase(config=input_name, show_progress=True)

    if mixdb.feature != feature_mode:
        logger.exception(f'Feature in mixture database does not match feature in model')
        raise SystemExit(1)

    mixids = mixdb.mixids_to_list(mixids)
    if reset:
        # reset mode cycles through each file one at a time
        for mixid in mixids:
            feature, _ = mixdb.mixture_ft(mixid)

            feature, predict = pad_and_predict(hypermodel=hypermodel,
                                               built_model=built_model,
                                               feature=feature,
                                               frames_per_batch=frames_per_batch)

            output_name = join(output_dir, mixdb.mixtures[mixid].name)
            with h5py.File(output_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset(name='predict', data=predict)
    else:
        # Run all data at once using a data generator
        feature = KerasFromH5(mixdb=mixdb,
                              mixids=mixids,
                              batch_size=hypermodel.batch_size,
                              timesteps=hypermodel.timesteps,
                              flatten=hypermodel.flatten,
                              add1ch=hypermodel.add1ch)

        predict = built_model.predict(feature, batch_size=hypermodel.batch_size, verbose=1)
        predict, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)

        # Write data to separate files
        for idx, mixid in enumerate(mixids):
            output_name = join(output_dir, mixdb.mixtures[mixid].name)
            with h5py.File(output_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset('predict', data=predict[feature.file_indices[idx]])

    logger.info(f'Saved results to {output_dir}')


def pad_and_predict(hypermodel: Any, built_model: Any, feature: Feature, frames_per_batch: int) -> (Feature, Predict):
    import numpy as np

    from sonusai.utils import reshape_inputs
    from sonusai.utils import reshape_outputs

    frames = feature.shape[0]
    padding = frames_per_batch - frames % frames_per_batch
    feature = np.pad(array=feature, pad_width=((0, padding), (0, 0), (0, 0)))
    feature, _ = reshape_inputs(feature=feature,
                                batch_size=hypermodel.batch_size,
                                timesteps=hypermodel.timesteps,
                                flatten=hypermodel.flatten,
                                add1ch=hypermodel.add1ch)
    predict = built_model.predict(feature, batch_size=hypermodel.batch_size, verbose=1)
    predict, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)
    predict = predict[:frames, :]
    return feature, predict


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



"""Map functions for parallel processing.

p_map:  Performs a parallel ordered map.
p_umap: Performs a parallel unordered map.
"""

from typing import Any
from typing import Callable
from typing import Generator
from typing import Iterable
from typing import List


def _parallel(ordered: bool, function: Callable, *iterables: Iterable, **kwargs: Any) -> Generator:
    """Returns a generator for a parallel map.

    Arguments:
        ordered     bool        True for an ordered map, False for an unordered map.
        function    Callable    The function to apply to each element of the given Iterable.
        iterables   Iterable    One or more Iterables containing the data to be mapped.

    Returns:
        A generator which will apply the function to each element of the given Iterables
        in parallel in order.
    """
    import multiprocessing as mp

    # Extract num_cpus
    num_cpus = kwargs.pop('num_cpus', None)

    # Determine num_cpus
    if num_cpus is None:
        num_cpus = mp.cpu_count()
    elif type(num_cpus) == float:
        num_cpus = int(round(num_cpus * mp.cpu_count()))

    # Create parallel generator
    initializer = kwargs.pop('initializer', None)
    initargs = kwargs.pop('initargs', None)
    map_type = 'imap' if ordered else 'imap_unordered'
    with mp.Pool(num_cpus, initializer=initializer, initargs=initargs) as pool:
        map_func = getattr(pool, map_type)

        for item in map_func(function, *iterables):
            yield item


def p_map(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a parallel ordered map."""
    return list(_parallel(True, function, *iterables, **kwargs))


def p_umap(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a parallel unordered map."""
    return list(_parallel(False, function, *iterables, **kwargs))



import numpy as np

from sonusai.mixture.types import Predict
from sonusai.mixture.types import Truth


def calculate_optimal_thresholds(truth: Truth,
                                 predict: Predict,
                                 timesteps: int = 0,
                                 truth_thr: float = 0.5) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    """Calculates optimal thresholds for each class from one-hot prediction and truth data where both are
    one-hot probabilities (or quantized decisions) with size [frames, num_classes] or [frames, timesteps, num_classes].

    Returns:
      thresholds_opt_pr     [num_classes, 1] optimal thresholds for PR-curve (F1) performance
      thresholds_opt_roc    [num_classes, 1] optimal thresholds for ROC-curve (TPR/FPR) performance
      AP                    [num_classes, 1]
      AUC                   [num_classes, 1]

    Optional truth_thr is the decision threshold(s) applied to truth one-hot input allowing truth to optionally be
    continuous probabilities. Default is 0.5.
    """
    from sklearn.metrics import average_precision_score
    from sklearn.metrics import precision_recall_curve
    from sklearn.metrics import roc_auc_score
    from sklearn.metrics import roc_curve

    from sonusai import SonusAIError
    from sonusai.utils import get_num_classes_from_predict
    from sonusai.utils import reshape_outputs

    if truth.shape != predict.shape:
        raise SonusAIError('truth and predict are not the same shape')

    predict, truth = reshape_outputs(predict=predict, truth=truth, timesteps=timesteps)
    num_classes = get_num_classes_from_predict(predict=predict, timesteps=timesteps)

    # Apply decision to truth input
    truth_binary = np.int8(truth >= truth_thr)

    AP = np.zeros((num_classes, 1))
    AUC = np.zeros((num_classes, 1))
    thresholds_opt_pr = np.zeros((num_classes, 1))
    thresholds_opt_roc = np.zeros((num_classes, 1))
    eps = np.finfo(float).eps
    for nci in range(num_classes):
        # Average Precision also called area under the PR curve AUCPR and
        # AUC ROC curve using binary-ized truth and continuous prediction probabilities
        # sklearn returns nan if no active truth in a class but w/un-suppressible div-by-zero warning
        if sum(truth_binary[:, nci]) == 0:  # no active truth must be NaN
            thresholds_opt_pr[nci] = np.NaN
            thresholds_opt_roc[nci] = np.NaN
            AUC[nci] = np.NaN
            AP[nci] = np.NaN
        else:
            AP[nci] = average_precision_score(truth_binary[:, nci], predict[:, nci], average=None)
            AUC[nci] = roc_auc_score(truth_binary[:, nci], predict[:, nci], average=None)

            # Optimal threshold from PR curve, optimizes f-score
            precision, recall, thrpr = precision_recall_curve(truth_binary[:, nci], predict[:, nci])
            fscore = (2 * precision * recall) / (precision + recall + eps)
            ix = np.argmax(fscore)  # index of largest f1 score
            thresholds_opt_pr[nci] = thrpr[ix]

            # Optimal threshold from ROC curve, optimizes J-statistic (TPR-FPR) or gmean
            fpr, tpr, thrroc = roc_curve(truth_binary[:, nci], predict[:, nci])
            # J = tpr - fpr                  # J can result in thr > 1
            gmeans = np.sqrt(tpr * (1 - fpr))  # gmean seems better behaved
            ix = np.argmax(gmeans)
            thresholds_opt_roc[nci] = thrroc[ix]

    return thresholds_opt_pr, thresholds_opt_roc, AP, AUC



import numpy as np

from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import GenMixData


def _initialize_target_audio(mixdb: MixtureDatabase,
                             mrecord: MRecord,
                             raw_target_audios: AudiosT) -> AudiosT:
    """ Apply augmentation and update target metadata
    """
    from sonusai.mixture import pad_audio_to_length
    from sonusai.mixture.augmentation import apply_augmentation

    targets = []
    mrecord.target_gain = []
    for idx in range(len(raw_target_audios)):
        target_augmentation = mixdb.target_augmentations[mrecord.target_augmentation_index[idx]]

        targets.append(apply_augmentation(audio=raw_target_audios[idx],
                                          augmentation=target_augmentation,
                                          length_common_denominator=mixdb.feature_step_samples))

        # target_gain is used to back out the gain augmentation in order to return the target audio
        # to its normalized level when calculating truth (if needed).
        if target_augmentation.gain is not None:
            target_gain = 10 ** (target_augmentation.gain / 20)
        else:
            target_gain = 1
        mrecord.target_gain.append(target_gain)

    mrecord.samples = max([len(item) for item in targets])

    for idx in range(len(targets)):
        targets[idx] = pad_audio_to_length(audio=targets[idx], length=mrecord.samples)

    return targets


def initialize_target(mixdb: MixtureDatabase,
                      mrecord: MRecord,
                      raw_target_audios: AudiosT) -> MRecord:
    """ Apply augmentation and update target metadata
    """
    _initialize_target_audio(mixdb, mrecord, raw_target_audios)
    return mrecord


def _initialize_mixture_gains(mrecord: MRecord,
                              target_audios: AudiosT,
                              noise_audio: AudioT) -> MRecord:
    from sonusai.utils import dbfs_to_sample

    target_audio = np.float32(sum(target_audios))

    if mrecord.snr < -96:
        # Special case for zeroing out target data
        mrecord.target_snr_gain = 0
        mrecord.noise_snr_gain = 1
        # Setting target_gain to zero will cause the truth to be all zeros.
        mrecord.target_gain = [0] * len(mrecord.target_gain)
    elif mrecord.snr > 96:
        # Special case for zeroing out noise data
        mrecord.target_snr_gain = 1
        mrecord.noise_snr_gain = 0
    else:
        target_energy = np.mean(np.square(target_audio))
        noise_energy = np.mean(np.square(np.float32(noise_audio)))
        if noise_energy == 0:
            noise_gain = 1
        else:
            noise_gain = np.sqrt(target_energy / noise_energy) / 10 ** (mrecord.snr / 20)

        # Check for noise_gain > 1 to avoid clipping
        if noise_gain > 1:
            mrecord.target_snr_gain = 1 / noise_gain
            mrecord.noise_snr_gain = 1
        else:
            mrecord.target_snr_gain = 1
            mrecord.noise_snr_gain = noise_gain

    # Check for clipping in mixture
    gain_adjusted_target_audio = target_audio * mrecord.target_snr_gain
    gain_adjusted_noise_audio = np.float32(noise_audio) * mrecord.noise_snr_gain
    mixture_audio = gain_adjusted_target_audio + gain_adjusted_noise_audio
    if any(abs(mixture_audio) >= 32768):
        # Clipping occurred; lower gains to bring audio within int16 bounds
        gain_adjustment = dbfs_to_sample(-0.25) / max(abs(mixture_audio))
        mrecord.target_snr_gain *= gain_adjustment
        mrecord.noise_snr_gain *= gain_adjustment

    return mrecord


def initialize_mixture(mixdb: MixtureDatabase,
                       mrecord: MRecord,
                       raw_target_audios: AudiosT,
                       noise_audio: AudioT) -> (MRecord, GenMixData):
    from sonusai.mixture import apply_gain
    from sonusai.mixture import apply_ir
    from sonusai.mixture import get_next_noise

    targets = _initialize_target_audio(mixdb=mixdb, mrecord=mrecord, raw_target_audios=raw_target_audios)
    noise = get_next_noise(audio=noise_audio, offset=mrecord.noise_offset, length=mrecord.samples)
    mrecord = _initialize_mixture_gains(mrecord=mrecord, target_audios=targets, noise_audio=noise)
    targets = [apply_gain(audio=target, gain=mrecord.target_snr_gain) for target in targets]
    noise = apply_gain(audio=noise, gain=mrecord.noise_snr_gain)

    # Apply impulse response to targets
    targets_ir = []
    for idx, target in enumerate(targets):
        ir_idx = mixdb.target_augmentations[mrecord.target_augmentation_index[idx]].ir
        if ir_idx is not None:
            targets_ir.append(apply_ir(audio=target, ir=mixdb.ir_data[ir_idx]))
        else:
            targets_ir.append(target)

    mixture = sum(targets_ir) + noise

    return mrecord, GenMixData(mixture=mixture, targets=targets, noise=noise)
